'''
Check call-throughs.
'''

import numbers
import numpy as np
import pytest
import scipy

import mqr

def test_test_1sample():
    x = np.array([0.88612542, 0.67327485, 0.8307321 , 0.46658372, 0.62968216,
        0.67137016, 0.14714929, 0.02982314, 0.45770049, 0.70468434,
        0.97795913, 0.78578141, 0.14394705, 0.73653195, 0.40538661,])
    H0_median = 0.5

    alternative = 'two-sided'
    method = 'sign'
    res = mqr.inference.nonparametric.median.test_1sample(x, H0_median, alternative, method)
    assert res.description == 'median'
    assert res.alternative == alternative
    assert res.method == method
    assert res.sample_stat == 'median(x)'
    assert res.sample_stat_target == H0_median
    assert isinstance(res.sample_stat_value, numbers.Number)
    assert isinstance(res.stat, numbers.Number)
    assert isinstance(res.pvalue, numbers.Number)

    alternative = 'less'
    method = 'wilcoxon'
    res = mqr.inference.nonparametric.median.test_1sample(x, H0_median, alternative, method)
    assert res.description == 'median'
    assert res.alternative == alternative
    assert res.method == method
    assert res.sample_stat == 'median(x)'
    assert res.sample_stat_target == H0_median
    assert isinstance(res.sample_stat_value, numbers.Number)
    assert isinstance(res.stat, numbers.Number)
    assert isinstance(res.pvalue, numbers.Number)

def test_test_nsample():
    x, y, z = np.array([
       [0.36522057, 0.37377119, 0.86150726, 0.96718967, 0.47966424,
        0.74979588, 0.35591634, 0.75272332, 0.17909715, 0.66216129],
       [0.30579918, 0.01280289, 0.2218879 , 0.77332064, 0.18087063,
        0.85439362, 0.41027609, 0.96153719, 0.94262441, 0.88509784],
       [0.1558813 , 0.25968253, 0.32619182, 0.89433078, 0.51197809,
        0.5814941 , 0.60785036, 0.32729355, 0.68637953, 0.9616885 ]])

    alternative = 'two-sided'
    method = 'kruskal-wallis'
    res = mqr.inference.nonparametric.median.test_nsample(x, y, z, alternative=alternative, method=method)
    assert res.description == 'equality of medians'
    assert res.alternative == alternative
    assert res.method == method
    assert res.sample_stat == 'median(x_i)'
    assert res.sample_stat_target == 'median(x_j)'
    assert np.isnan(res.sample_stat_value)
    assert isinstance(res.stat, numbers.Number)
    assert isinstance(res.pvalue, numbers.Number)

    alternative = 'greater'
    method = 'mann-whitney'
    res = mqr.inference.nonparametric.median.test_nsample(x, y, alternative=alternative, method=method)
    assert res.description == 'equality of medians'
    assert res.alternative == alternative
    assert res.method == method
    assert res.sample_stat == 'median(x) - median(y)'
    assert res.sample_stat_target == 0.0
    assert np.isnan(res.sample_stat_value)
    assert isinstance(res.stat, numbers.Number)
    assert isinstance(res.pvalue, numbers.Number)



import mqr

import numpy as np
import pandas as pd
import pyDOE3
import pytest

def test_Design_replicate():
    names = ['a', 'b']
    levels = [2, 2]
    design = (
        mqr.doe.Design.from_fullfact(names, levels).as_block(1) +
        mqr.doe.Design.from_centrepoints(names, 2).as_block(2))

    replicated = design.replicate(2)
    assert np.all(replicated.runs.values[0::2] == replicated.runs.values[1::2] - 1)
    assert np.all(replicated.pttypes.values[0::2] == replicated.pttypes.values[1::2])
    assert np.all(replicated.levels.values[0::2] == replicated.levels.values[1::2])

    replicated = design.replicate(2, 'Rep')
    assert np.all(replicated.runs.values[0::2] == replicated.runs.values[1::2] - 1)
    assert np.all(replicated.pttypes.values[0::2] == replicated.pttypes.values[1::2])
    assert np.all(replicated.levels.values[0::2] == replicated.levels.values[1::2])
    assert np.all(replicated.blocks['Rep'][0::2] == 1)
    assert np.all(replicated.blocks['Rep'][1::2] == 2)

    levels = [2, 3]
    design = mqr.doe.Design.from_fullfact(names, levels)

    replicated = design.replicate(2)
    assert np.all(replicated.runs.values[0::2] == replicated.runs.values[1::2] - 1)
    assert np.all(replicated.levels.values[0::2] == replicated.levels.values[1::2])

def test_Design_as_block():
    names = ['a', 'b']
    levels = [2, 2]
    design = (
        mqr.doe.Design.from_fullfact(names, levels).as_block(1) +
        mqr.doe.Design.from_centrepoints(names, 2).as_block(2))

    assert np.all(design.as_block(4).blocks == 4)

def test_Design_to_df():
    names = ['a', 'b']
    levels = [2, 2]
    design = (
        mqr.doe.Design.from_fullfact(names, levels).as_block(1) +
        mqr.doe.Design.from_centrepoints(names, 2).as_block(2))
    df = design.to_df()

    assert np.all(df.index == design.runs)
    assert np.all(df['PtType'] == design.pttypes)
    assert np.all(df[['Block']] == design.blocks)
    assert np.all(df.loc[:, 'a':'b'] == design.levels)

def test_Design_get_factor_df():
    names = ['a', 'b']
    levels = [2, 2]
    design = (
        mqr.doe.Design.from_fullfact(names, levels).as_block(1) +
        mqr.doe.Design.from_centrepoints(names, 2).as_block(2))
    factor_df = design.get_factor_df('a', ref_levels=1.23)

    assert list(factor_df['a']) == sorted(design.levels['a'].unique())
    assert all(factor_df['b'] == 1.23)

def test_Design_randomise_runs():
    names = ['a', 'b', 'c']
    levels = [2, 2, 3]
    design = mqr.doe.Design.from_fullfact(names, levels)
    df = design.to_df()

    # complete randomisation
    np.random.seed(0)
    rdesign = design.randomise_runs()
    rdf = rdesign.to_df()

    assert set(design.runs) == set(rdesign.runs)
    assert list(design.runs) != list(rdesign.runs) # True when seeded with 0
    for run in design.runs:
        assert list(df.loc[run]) == list(rdf.loc[run])

    # one level ordering
    np.random.seed(0)
    rdesign = design.randomise_runs('a')
    rdf = rdesign.to_df()

    assert set(design.runs) == set(rdesign.runs)
    assert sorted(rdf['a']) == list(rdf['a'])
    assert list(design.runs) != list(rdesign.runs) # True when seeded with 0
    for run in design.runs:
        assert list(df.loc[run]) == list(rdf.loc[run])

    # two level ordering
    np.random.seed(0)
    rdesign = design.randomise_runs('a', 'b')
    rdf = rdesign.to_df()

    assert set(design.runs) == set(rdesign.runs)
    assert sorted(rdf['a']) == list(rdf['a'])
    for level in rdf['a'].unique():
        bs = rdf.query(f'a == {level}')['b']
        print(bs)
        assert sorted(bs) == list(bs)
    assert list(design.runs) != list(rdesign.runs) # True when seeded with 0
    for run in design.runs:
        assert list(df.loc[run]) == list(rdf.loc[run])

    # complete randomisation
    np.random.seed(0)
    design.pttypes = pd.Series(1, index=design.runs)
    df = design.to_df()
    rdesign = design.randomise_runs()
    rdf = rdesign.to_df()

    assert set(design.runs) == set(rdesign.runs)
    assert list(design.runs) != list(rdesign.runs) # True when seeded with 0
    for run in design.runs:
        assert list(df.loc[run]) == list(rdf.loc[run])

def test_Design_add():
    names = ['a', 'b']
    levels = [2, 2]
    design1 = mqr.doe.Design.from_fullfact(names, levels).as_block(1)
    design2 = mqr.doe.Design.from_centrepoints(names, 3).as_block(2)

    design = design1 + design2
    assert list(design.runs) == list(design1.runs) + list(design2.runs + len(design1))
    assert list(design.pttypes) == list(design1.pttypes) + list(design2.pttypes)
    assert np.all(design.levels.values == pd.concat([design1.levels, design2.levels]).values)
    assert np.all(design.blocks.values == pd.concat([design1.blocks, design2.blocks]).values)

    design1.pttypes = None
    design = design1 + design2
    assert list(design.runs) == list(design1.runs) + list(design2.runs + len(design1))
    assert np.all(np.isnan(design.pttypes[:len(design1)]))
    assert list(design.pttypes[len(design1):] == list(design2.pttypes))
    assert np.all(design.levels.values == pd.concat([design1.levels, design2.levels]).values)
    assert np.all(design.blocks.values == pd.concat([design1.blocks, design2.blocks]).values)

    design1 = mqr.doe.Design.from_fullfact(names, levels).as_block(1)
    design2.pttypes = None
    design = design1 + design2
    assert list(design.runs) == list(design1.runs) + list(design2.runs + len(design1))
    assert list(design.pttypes[:len(design1)] == list(design1.pttypes))
    assert np.all(np.isnan(design.pttypes[len(design1):]))
    assert np.all(design.levels.values == pd.concat([design1.levels, design2.levels]).values)
    assert np.all(design.blocks.values == pd.concat([design1.blocks, design2.blocks]).values)

    design1.pttypes = None
    design2.pttypes = None
    design = design1 + design2
    assert list(design.runs) == list(design1.runs) + list(design2.runs + len(design1))
    assert design.pttypes == None
    assert np.all(design.levels.values == pd.concat([design1.levels, design2.levels]).values)
    assert np.all(design.blocks.values == pd.concat([design1.blocks, design2.blocks]).values)

def test_Design_transform():
    names = ['a', 'b', 'c']
    levels = [2, 2, 2]
    design = mqr.doe.Design.from_fullfact(names, levels).as_block(2)
    
    tr = design.transform(
        a={-1: 5, 1: 15},
        c=lambda x: 2*x)
    exp_levels = np.array([
        [5, 15, 5, 15, 5, 15, 5, 15],
        [-1, -1, 1, 1, -1, -1, 1, 1],
        [-2, -2, -2, -2, 2, 2, 2, 2],
    ]).T

    assert list(tr.runs) == list(design.runs)
    assert list(tr.pttypes) == list(design.pttypes)
    assert np.all(tr.blocks == design.blocks)
    assert np.all(tr.levels.values == exp_levels)

def test_Design_from_levels():
    names = ['a', 'b', 'c']
    levels = np.array([
        [-1, -1, 1, 1],
        [-1, 1, -1, 1],
        [1, -1, -1, 1],
    ]).T
    runs = pd.Index([3, 4, 5, 6])
    
    design = mqr.doe.Design.from_levels(names, levels, runs)

    assert list(design.runs) == list(runs)
    assert np.all(design.levels.values == levels)

def test_Design_from_fullfact():
    names = ['a', 'b']
    levels = [2, 2]    
    exp_runs = list(range(1, 2 * 2 + 1))
    exp_levels = pyDOE3.fullfact(levels)

    design = mqr.doe.Design.from_fullfact(names, levels, scale_origin=True, pttypes=True)
    assert list(design.runs) == exp_runs
    assert np.all(design.pttypes == 1)
    assert np.all(design.levels == exp_levels * 2 - 1)

    design = mqr.doe.Design.from_fullfact(names, levels, scale_origin=True, pttypes=False)
    assert list(design.runs) == exp_runs
    assert design.pttypes == None
    assert np.all(design.levels == exp_levels * 2 - 1)

    design = mqr.doe.Design.from_fullfact(names, levels, scale_origin=False, pttypes=True)
    assert list(design.runs) == exp_runs
    assert np.all(design.pttypes == 1)
    assert np.all(design.levels == exp_levels)

    design = mqr.doe.Design.from_fullfact(names, levels, scale_origin=False, pttypes=False)
    assert list(design.runs) == exp_runs
    assert np.all(design.pttypes == None)
    assert np.all(design.levels == exp_levels)

def test_Design_from_fracfact():
    names = ['a', 'b', 'c']
    gen = 'a b ab'
    exp_runs = list(range(1, 2 * 2 + 1))
    exp_levels = pyDOE3.fracfact(gen)

    design = mqr.doe.Design.from_fracfact(names, gen)
    assert list(design.runs) == exp_runs
    assert np.all(design.pttypes == 1)
    assert np.all(design.levels == exp_levels)

def test_Design_from_ccdesign():
    names = ['a', 'b', 'c']
    exp_runs = list(range(1, 2 ** 3 + 3 * 2 + 1))
    exp_levels = pyDOE3.ccdesign(len(names), center=(0, 0))

    design = mqr.doe.Design.from_ccdesign(names)
    assert list(design.runs) == exp_runs
    assert np.all(design.pttypes[:8] == 1)
    assert np.all(design.pttypes[8:] == 2)
    assert np.all(np.isclose(design.levels, exp_levels))

def test_Design_from_centrepoints():
    names = ['a', 'b', 'c']
    exp_runs = list(range(1, 5))

    design = mqr.doe.Design.from_centrepoints(names, 4)
    assert list(design.runs) == exp_runs
    assert np.all(design.pttypes == 0)
    assert np.all(design.levels == 0)

def test_Design_from_axial():
    names = ['a', 'b', 'c']
    exp_runs = list(range(1, 5))
    exp_levels = pyDOE3.ccdesign(len(names), center=(0, 0))[8:]

    design = mqr.doe.Design.from_axial(names, exclude=['b'])
    assert list(design.runs) == exp_runs
    assert list(design.pttypes == 2)
    assert np.all(design.levels['a'] == np.sqrt(3) * np.array([-1, 1, 0, 0]))
    assert np.all(design.levels['b'] == 0)
    assert np.all(design.levels['c'] == np.sqrt(3) * np.array([0, 0, -1, 1]))

    design = mqr.doe.Design.from_axial(names, exclude=['b'], magnitude=2)
    assert list(design.runs) == exp_runs
    assert list(design.pttypes == 2)
    assert np.all(design.levels['a'] == 2 * np.array([-1, 1, 0, 0]))
    assert np.all(design.levels['b'] == 0)
    assert np.all(design.levels['c'] == 2 * np.array([0, 0, -1, 1]))

    design = mqr.doe.Design.from_axial(names)
    exp_runs = list(range(1, 7))
    assert list(design.runs) == exp_runs
    assert list(design.pttypes == 2)
    assert np.all(design.levels['a'] == np.sqrt(3) * np.array([-1, 1, 0, 0, 0, 0]))
    assert np.all(design.levels['b'] == np.sqrt(3) * np.array([0, 0, -1, 1, 0, 0]))
    assert np.all(design.levels['c'] == np.sqrt(3) * np.array([0, 0, 0, 0, -1, 1]))

def test_Design_is_centre():
    assert mqr.doe.Design._is_centre(pd.Series([0, 0, 0, 0]))
    assert not mqr.doe.Design._is_centre(pd.Series([0, 0, 1]))
    assert not mqr.doe.Design._is_centre(pd.Series([1, 1, 1]))
    assert mqr.doe.Design._is_centre(pd.Series([0]))
    assert not mqr.doe.Design._is_centre(pd.Series([1]))

def test_Design_is_corner():
    assert mqr.doe.Design._is_corner(pd.Series([1, -1, -1, 1]))
    assert not mqr.doe.Design._is_corner(pd.Series([1, -1, -1, 0]))
    assert not mqr.doe.Design._is_corner(pd.Series([0, 0, 0]))
    assert not mqr.doe.Design._is_corner(pd.Series([0]))
    assert mqr.doe.Design._is_corner(pd.Series([1]))

def test_Design_is_axial():
    assert mqr.doe.Design._is_axial(pd.Series([0, 1, 0, 0]))
    assert mqr.doe.Design._is_axial(pd.Series([0, 0, -1, 0]))
    assert not mqr.doe.Design._is_axial(pd.Series([0, 1, -1, 0]))
    assert not mqr.doe.Design._is_axial(pd.Series([0, 1, -1, 0]))
    assert not mqr.doe.Design._is_axial(pd.Series([0]))
    assert mqr.doe.Design._is_axial(pd.Series([1]))

def test_Design_pttype():
    assert mqr.doe.Design._pttype(pd.Series([0, 0, 0])) == 0
    assert mqr.doe.Design._pttype(pd.Series([-1, 1, 1])) == 1
    assert mqr.doe.Design._pttype(pd.Series([-1, 0, 0])) == 2
    assert mqr.doe.Design._pttype(pd.Series([-1])) == 1 # corner takes precedence when only one dimension
    assert mqr.doe.Design._pttype(pd.Series([0])) == 0
    assert mqr.doe.Design._pttype(pd.Series([1, 0, -1])) is None

def test_scale():
    assert list(mqr.doe.Design._scale(2)(np.array([0, 1]))) == list([-1, 1])
    assert list(mqr.doe.Design._scale(3)(np.array([0, 1, 2]))) == list([-1, 0, 1])
    assert list(mqr.doe.Design._scale(4)(np.array([0, 1, 2, 3]))) == list([-3, -1, 1, 3])
    assert list(mqr.doe.Design._scale(5)(np.array([0, 1, 2, 3, 4]))) == list([-2, -1, 0, 1, 2])

def test_Transform_from_map():
    assert list(mqr.doe.Transform.from_map({1: 4, 2: 8})(np.array([0, 1, 2]))) == list([0, 4, 8])
    assert list(mqr.doe.Transform.from_map({0: 4, 2: 8})(np.array([0, 1, 2]))) == list([4, 6, 8])
    assert list(mqr.doe.Transform.from_map({1: 4, 4: -8})(np.array([0, 1, 2]))) == list([8, 4, 0])



from mqr.inference.hyptest import HypothesisTest

import mqr.inference.lib.util as util

def test(x, y, alternative='two-sided', method='spearman'):
    """
    Hypothesis test for correlation between two samples.

    Null hypothesis
        corr(`x`, `y`) == 0

    Parameters
    ----------
    x, y : array_like
        Test correlation of these two equal-length samples.
    alternative : {'two-sided', 'less', 'greater'}, optional
        Sense of alternative hypothesis.
    method : {'spearman', 'kendall'}, optional
        | 'spearman'
        |   Spearman correlation coefficient.
            Calls :func:`scipy..spearmanr <scipy.stats.spearmanr>`.
        | 'kendall'
        |   Kendall's tau measure.
            Calls :func:`scipy..kendalltau <scipy.stats.kendalltau>`.

    Returns
    -------
    :class:`mqr.inference.hyptest.HypothesisTest`
    """
    import numpy as np
    import scipy.stats as st
    import mqr

    if method == 'spearman':
        stat, pvalue = st.spearmanr(
            a=x,
            b=y,
            alternative=alternative)
    elif method == 'kendall':
        stat, pvalue = st.kendalltau(
            x=x,
            y=y,
            alternative=alternative)
    else:
        raise ValueError(util.method_error_msg(method, ['spearman', 'kendall']))

    x_name = util.var_name(x, 'x')
    y_name = util.var_name(y, 'y')
    return HypothesisTest(
        description='correlation coefficient',
        alternative=alternative,
        method=method,
        sample_stat=f'corr({x_name}, {y_name})',
        sample_stat_target=0.0,
        sample_stat_value=stat,
        stat=stat,
        pvalue=pvalue,)



'''
Check call-throughs.
'''

import numbers
import numpy as np
import pytest
import scipy

import mqr

def test_confint():
    x = np.array([1, 2, 4, 5])
    y = np.array([1, 3, 4, 4])
    conf = 0.90

    bounded = 'both'
    res = mqr.inference.correlation.confint(x, y, conf, bounded)
    assert res.name == 'correlation'
    assert res.method == 'fisher-z'
    assert res.value == pytest.approx(0.9036961141150641)
    assert res.lower == pytest.approx(-0.151652753)
    assert res.upper == pytest.approx(0.996236807)
    assert res.conf == conf

    bounded = 'above'
    res = mqr.inference.correlation.confint(x, y, conf, bounded)
    assert res.name == 'correlation'
    assert res.method == 'fisher-z'
    assert res.value == pytest.approx(0.9036961141150641)
    assert res.lower == -1.0
    assert res.upper == pytest.approx(0.9922332)
    assert res.conf == conf

    bounded = 'below'
    res = mqr.inference.correlation.confint(x, y, conf, bounded)
    assert res.name == 'correlation'
    assert res.method == 'fisher-z'
    assert res.value == pytest.approx(0.9036961141150641)
    assert res.lower == pytest.approx(0.2074167)
    assert res.upper == 1.0
    assert res.conf == conf

def test_test():
    x = np.array([1, 2, 4, 5])
    y = np.array([1, 3, 4, 4])

    H0_corr = 0.0
    alternative = 'two-sided'
    res = mqr.inference.correlation.test(x, y, H0_corr, alternative)
    assert res.description == 'correlation coefficient'
    assert res.alternative == alternative
    assert res.method == 'pearson'
    assert res.sample_stat == 'corr(x, y)'
    assert res.sample_stat_target == H0_corr
    assert res.sample_stat_value == pytest.approx(0.9037, abs=1e-4)
    assert res.stat == pytest.approx(2.9848, abs=1e-4)
    assert res.pvalue == pytest.approx(0.0963, abs=1e-4)

    H0_corr = 0.0
    alternative = 'less'
    res = mqr.inference.correlation.test(x, y, H0_corr, alternative)
    assert res.description == 'correlation coefficient'
    assert res.alternative == alternative
    assert res.method == 'pearson'
    assert res.sample_stat == 'corr(x, y)'
    assert res.sample_stat_target == H0_corr
    assert res.sample_stat_value == pytest.approx(0.9037, abs=1e-4)
    assert res.stat == pytest.approx(2.9848, abs=1e-4)
    assert res.pvalue == pytest.approx(0.9518, abs=1e-4)

    H0_corr = 0.0
    alternative = 'greater'
    res = mqr.inference.correlation.test(x, y, H0_corr, alternative)
    assert res.description == 'correlation coefficient'
    assert res.alternative == alternative
    assert res.method == 'pearson'
    assert res.sample_stat == 'corr(x, y)'
    assert res.sample_stat_target == H0_corr
    assert res.sample_stat_value == pytest.approx(0.9037, abs=1e-4)
    assert res.stat == pytest.approx(2.9848, abs=1e-4)
    assert res.pvalue == pytest.approx(0.04815, abs=1e-4)

    H0_corr = 0.4
    alternative = 'two-sided'
    res = mqr.inference.correlation.test(x, y, H0_corr, alternative)
    assert res.description == 'correlation coefficient'
    assert res.alternative == alternative
    assert res.method == 'fisher-z'
    assert res.sample_stat == 'corr(x, y)'
    assert res.sample_stat_target == H0_corr
    assert res.sample_stat_value == pytest.approx(0.9037, abs=1e-4)
    assert res.stat == pytest.approx(1.4920, abs=1e-4)
    assert res.pvalue == pytest.approx(0.2853, abs=1e-4)

    H0_corr = 0.4
    alternative = 'less'
    res = mqr.inference.correlation.test(x, y, H0_corr, alternative)
    assert res.description == 'correlation coefficient'
    assert res.alternative == alternative
    assert res.method == 'fisher-z'
    assert res.sample_stat == 'corr(x, y)'
    assert res.sample_stat_target == H0_corr
    assert res.sample_stat_value == pytest.approx(0.9037, abs=1e-4)
    assert res.stat == pytest.approx(1.4920, abs=1e-4)
    assert res.pvalue == pytest.approx(0.8573, abs=1e-4)

    H0_corr = 0.4
    alternative = 'greater'
    res = mqr.inference.correlation.test(x, y, H0_corr, alternative)
    assert res.description == 'correlation coefficient'
    assert res.alternative == alternative
    assert res.method == 'fisher-z'
    assert res.sample_stat == 'corr(x, y)'
    assert res.sample_stat_target == H0_corr
    assert res.sample_stat_value == pytest.approx(0.9037, abs=1e-4)
    assert res.stat == pytest.approx(1.4920, abs=1e-4)
    assert res.pvalue == pytest.approx(0.1427, abs=1e-4)

def test_test_diff():
    x1 = np.array([1, 2, 4, 5])
    y1 = np.array([1, 3, 4, 4])
    x2 = np.array([1, 2, 4, 5])
    y2 = np.array([3, 3, 2, 0])
    H0_corr1 = 0.8
    H0_corr2 = -0.1

    alternative = 'two-sided'
    res = mqr.inference.correlation.test_diff(x1, y1, x2, y2, H0_corr1, H0_corr2, alternative)
    assert res.description == 'difference between correlation coefficients'
    assert res.alternative == alternative
    assert res.method == 'fisher-z'
    assert res.sample_stat == 'corr(x1, y1) - corr(x2, y2)'
    assert res.sample_stat_target == H0_corr1 - H0_corr2
    assert res.sample_stat_value == pytest.approx(1.8074, abs=1e-4)
    assert res.stat == pytest.approx(2.9840, abs=1e-4)
    assert res.pvalue == pytest.approx(0.2068, abs=1e-4)

    alternative = 'less'
    res = mqr.inference.correlation.test_diff(x1, y1, x2, y2, H0_corr1, H0_corr2, alternative)
    assert res.description == 'difference between correlation coefficients'
    assert res.alternative == alternative
    assert res.method == 'fisher-z'
    assert res.sample_stat == 'corr(x1, y1) - corr(x2, y2)'
    assert res.sample_stat_target == H0_corr1 - H0_corr2
    assert res.sample_stat_value == pytest.approx(1.8074, abs=1e-4)
    assert res.stat == pytest.approx(2.9840, abs=1e-4)
    assert res.pvalue == pytest.approx(0.8966, abs=1e-4)

    alternative = 'greater'
    res = mqr.inference.correlation.test_diff(x1, y1, x2, y2, H0_corr1, H0_corr2, alternative)
    assert res.description == 'difference between correlation coefficients'
    assert res.alternative == alternative
    assert res.method == 'fisher-z'
    assert res.sample_stat == 'corr(x1, y1) - corr(x2, y2)'
    assert res.sample_stat_target == H0_corr1 - H0_corr2
    assert res.sample_stat_value == pytest.approx(1.8074, abs=1e-4)
    assert res.stat == pytest.approx(2.9840, abs=1e-4)
    assert res.pvalue == pytest.approx(0.1034, abs=1e-4)



"""
===================================================
Process summary and capability (:mod:`mqr.process`)
===================================================

.. currentmodule:: mqr.process

User guide
    :doc:`/user_guide/summary-capability`

Detailed examples
    https://github.com/nklsxn/mqrpy-guide

Routines for summarising processes and their capability.

See `doc:mqr.process` for more details.

.. rubric:: Classes

.. autosummary::
    :toctree: generated/

    Sample
    Specification
    Capability
    Summary
"""

from dataclasses import dataclass, field
import matplotlib.pyplot as plt
import matplotlib.transforms as transforms
import numpy as np
import pandas as pd
import scipy
import scipy.stats as st
import seaborn as sns
from statsmodels.stats.diagnostic import normal_ad, kstest_normal

import mqr

@dataclass
class Sample:
    """
    Data and descriptive statistics for a single sample from a process.

    Construct using a pandas Series (ie. a column from a dataframe). Intended
    for use by a `Study` object.

    Attributes
    ----------
    name : str
        Name of the KPI or measurement.
    conf : float
        Confidence level to use in confidence intervals.
    data : pd.Series
        Sample measurements.

    ad_stat : float
        Anderson-Darling normality test statistic.
    ad_pvalue : float
        p-value associated with `ad_stat`.
    ks_stat : float
        Kolmogorov-Smirnov goodness of fit (with normal) test statistic.
    ks_pvalue : float
        p-value associated with `ks_stat`.

    nobs : int
        Number of measurements in the sample.
    mean : float
        Sample mean.
    sem : float
        Standard error of mean.
    std : float
        Sample standard deviation.
    var : float
        Sample variance.
    skewness : float
        Skewness.
    kurtosis : float
        Kurtosis.
    minimum : float
        Smallest observation.
    quartile1 : float
        25th percentile observation.
    median : float
        Median observation.
    quartile3 : float
        75th percentile obsevation.
    maximum : float
         Largest observation.
    iqr : float
        Inter-quartile range.

    conf_mean : ConfidenceInterval
        Conf interval on the mean.
    conf_var : ConfidenceInterval
        Conf interval on the variance.
    conf_quartile1 : ConfidenceInterval
        Conf interval on the 25th percentile.
    conf_median : ConfidenceInterval
        Conf interval on the median.
    conf_quartile3 : ConfidenceInterval
        Conf interval on the 75th percentile.
    outliers : array_like
        List of points falling further from a quartile than `1.5 * iqr`.

    Examples
    --------
    In a jupyter notebook, sample summaries are shown as HTML tables:

    >>> data = pd.read_csv(mqr.sample_data('study-random-5x5.csv'))
    >>> mqr.process.Sample(data['KPI1'])

    produces

    +--------------------------------+
    | KPI1                           |
    +================================+
    | Normality (Anderson-Darling).  |
    +--------------+-----------------+
    | Stat         | 0.34261         |
    +--------------+-----------------+
    | P-value      | 0.48588         |
    +--------------+-----------------+
    |                                |
    +--------------+-----------------+
    | N            | 120             |
    +--------------+-----------------+
    |                                |
    +--------------+-----------------+
    | Mean         | 149.97          |
    +--------------+-----------------+
    | StdDev       | 1.1734          |
    +--------------+-----------------+
    | Variance     | 1.3768          |
    +--------------+-----------------+
    | Skewness     | 0.23653         |
    +--------------+-----------------+
    |                                |
    +--------------+-----------------+
    | Kurtosis     | 0.34012         |
    +--------------+-----------------+
    | Minimum      | 147.03          |
    +--------------+-----------------+
    | 1st Quartile | 149.22          |
    +--------------+-----------------+
    | Median       | 149.97          |
    +--------------+-----------------+
    | 3rd Quartile | 150.56          |
    +--------------+-----------------+
    | Maximum      | 153.27          |
    +--------------+-----------------+
    |                                |
    +--------------+-----------------+
    | N Outliers.  | 5               |
    +--------------+-----------------+

    """
    name: str = None
    conf: float = field(default=np.nan, repr=False)
    data: pd.Series = field(default=None, repr=False)

    ad_stat: float = field(default=np.nan, repr=False)
    ad_pvalue: float = field(default=np.nan, repr=False)
    ks_stat: float = field(default=np.nan, repr=False)
    ks_pvalue: float = field(default=np.nan, repr=False)

    nobs: int = 0
    mean: np.float64 = np.nan
    sem: np.float64 = field(default=np.nan, repr=False)
    std: np.float64 = field(default=np.nan, repr=False)
    var: np.float64 = np.nan
    skewness: np.float64 = np.nan
    kurtosis: np.float64 = np.nan
    minimum: np.float64 = np.nan
    quartile1: np.float64 = np.nan
    median: np.float64 = np.nan
    quartile3: np.float64 = np.nan
    maximum: np.float64 = np.nan
    iqr: np.float64 = field(default=np.nan, repr=False)

    conf_mean: mqr.inference.confint.ConfidenceInterval = field(default=None, repr=False)
    conf_std: mqr.inference.confint.ConfidenceInterval = field(default=None, repr=False)
    conf_var: mqr.inference.confint.ConfidenceInterval = field(default=None, repr=False)
    conf_quartile1: mqr.inference.confint.ConfidenceInterval = field(default=None, repr=False)
    conf_median: mqr.inference.confint.ConfidenceInterval = field(default=None, repr=False)
    conf_quartile3: mqr.inference.confint.ConfidenceInterval = field(default=None, repr=False)

    outliers: np.ndarray = field(default=None, repr=False)

    def __init__(self, data, conf=0.95, ddof=1, name=None, num_display_fmt='#.5g'):
        import scipy.stats as st

        if hasattr(data, 'name'):
            self.name = data.name
        elif name is not None:
            self.name = name
        else:
            self.name = 'data'
        self.conf = conf
        self.data = data

        (self.ad_stat, self.ad_pvalue) = normal_ad(data)
        (self.ks_stat, self.ks_pvalue) = kstest_normal(data)

        self.nobs = len(data)
        self.mean = np.mean(data)
        self.sem = st.sem(data)
        self.std = np.std(data, ddof=ddof)
        self.var = np.var(data, ddof=ddof)
        self.skewness = st.skew(data)
        self.kurtosis = st.kurtosis(data)
        self.minimum = np.min(data)
        self.quartile1 = np.quantile(data, 0.25)
        self.median = np.median(data)
        self.quartile3 = np.quantile(data, 0.75)
        self.maximum = np.max(data)
        self.iqr = self.quartile3 - self.quartile1

        self.conf_mean = mqr.inference.mean.confint_1sample(data, conf=conf)
        self.conf_std = mqr.inference.stddev.confint_1sample(data, conf=conf)
        self.conf_var = mqr.inference.variance.confint_1sample(data, conf=conf)

        self.conf_quartile1 = mqr.inference.nonparametric.quantile.confint_1sample(data, q=0.25, conf=conf)
        self.conf_median = mqr.inference.nonparametric.quantile.confint_1sample(data, q=0.5, conf=conf)
        self.conf_quartile3 = mqr.inference.nonparametric.quantile.confint_1sample(data, q=0.75, conf=conf)

        self.outliers = np.concatenate([
            data[data<self.quartile1-1.5*self.iqr],
            data[data>self.quartile3+1.5*self.iqr]])

@dataclass
class Specification:
    """
    Process specification.

    Attributes
    ----------
    target : float
        Design value for the process.
    lsl : float
        Lower specification limit.
    usl : float
        Upper specification limit.
    """
    target: float
    lsl: float
    usl: float

@dataclass
class Capability:
    """
    Process capability values.

    Attributes
    ----------
    sample : Sample
        Sample that has this capability.
    spec : Specification
        Specification for which `sample` has this capability.
    cp : float
        Process potential. The capability of the process if it was centred at
        `Specification.target`.
    cpk : float
        Process capability. The number of standard deviations of process variation
        that fit in the specification, normalised by 3*sigma. Ie. a 6-sigma
        process has capability 2.0.
    defects_st : float
        Short-term defect rate, based on a fitted normal distribution.
    defects_lt : float
        Long-term defect rate, based on a normal distribution with 1.5*stddev
        larger than short-term.

    Examples
    --------
    Construct this object with a sample and a `mqr.process.Specification`:

    .. code-block:: python
        :emphasize-lines: 3

        data = pd.read_csv(mqr.sample_data('study-random-5x5.csv'))
        summary = mqr.process.Summary(data['KPI1'])
        spec = mqr.process.Specification(150, 147, 153)
        mqr.process.Capability(summary['KPI1'], spec)

    In a jupyter notebook, this produces the HTML table below. Iterables of
    `Capability` are shown as the same table with multiple columns.

    +-----------------+----------+
    |                 | KPI1     |
    +=================+==========+
    | USL             | 153.     |
    +-----------------+----------+
    | Target          | 150.     |
    +-----------------+----------+
    | LSL             | 147.     |
    +-----------------+----------+
    |                            |
    +-----------------+----------+
    | Cpk             | 0.844    |
    +-----------------+----------+
    | Cp              | 0.852    |
    +-----------------+----------+
    | Defectsst (ppm) | 1.06e+04 |
    +-----------------+----------+
    | Defectslt (ppm) | 8.83e+04 |
    +-----------------+----------+

    """
    sample: Sample
    spec: Specification

    cp: float
    cpk: float
    defects_st: float
    defects_lt: float

    def __init__(self, sample: Sample, spec: Specification):
        """
        Construct Capability.

        Attributes
        ----------
        sample : mqr.process.Sample
            Set of measurements from KPI.
        spec : mqr.process.Specification
            Specificatino for KPI.

        """
        self.sample = sample
        self.spec = spec

        self.cp = (spec.usl - spec.lsl) / (6 * sample.std)
        self.cpk = np.minimum(spec.usl - sample.mean, sample.mean - spec.lsl) / (3 * sample.std)
        in_spec = np.logical_and(sample.data >= spec.lsl, sample.data <= spec.usl)
        dist = st.norm(sample.mean, sample.std)
        dist_lt = st.norm(sample.mean, 1.5 * sample.std)
        self.defects_st = 1 - (dist.cdf(spec.usl) - dist.cdf(spec.lsl))
        self.defects_lt = 1 - (dist_lt.cdf(spec.usl + 1.5 * sample.std) -
                               dist_lt.cdf(spec.lsl - 1.5 * sample.std))

@dataclass
class Summary:
    """
    Measurements and summary statistics for a set of samples from a process.

    Attributes
    ----------
    data : pd.DataFrame
        Measurements with KPIs in each column, and possibly other columns like
        run lables, operator IDs, etc.
    samples : dict[str, mqr.process.Sample]
        Automatically constructed. Dict of `mqr.process.Sample` for each sample
        in data.
    capabilities : dict[str, Capability]
        Automatically constructed when initialised with `Specification`s. Dict
        of `mqr.process.Capability` for each sample in data.

    Examples
    --------
    Construct this object using a dataframe of measurements, optionally providing
    a list of columns to include:

    >>> data = pd.read_csv(mqr.sample_data('study-random-5x5.csv'))
    >>> mqr.process.Study(data)

    That input is shown in notebooks as an HTML table:

    +--------------+---------+-----------+-----------+----------+----------+
    |              | KPI1    | KPI2      | KPI3      | KPO1     | KPO2     |
    +==============+=========+===========+===========+==========+==========+
    | Normality (Anderson-Darling)                                         |
    +--------------+---------+-----------+-----------+----------+----------+
    | Stat         | 0.34261 | 0.23796   | 1.1874    | 0.19203  | 0.70213  |
    +--------------+---------+-----------+-----------+----------+----------+
    | P-value      | 0.48588 | 0.77835   | 0.0040775 | 0.89417  | 0.065144 |
    +--------------+---------+-----------+-----------+----------+----------+
    +--------------+---------+-----------+-----------+----------+----------+
    | N            | 120     | 120       | 120       | 120      | 120      |
    +--------------+---------+-----------+-----------+----------+----------+
    +--------------+---------+-----------+-----------+----------+----------+
    | Mean         | 149.97  | 20.003    | 14.004    | 160.05   | 4.0189   |
    +--------------+---------+-----------+-----------+----------+----------+
    | StdDev       | 1.1734  | 0.24527   | 0.75643   | 2.0489   | 1.5634   |
    +--------------+---------+-----------+-----------+----------+----------+
    | Variance     | 1.3768  | 0.060156  | 0.57219   | 4.1979   | 2.4443   |
    +--------------+---------+-----------+-----------+----------+----------+
    | Skewness     | 0.23653 | -0.31780  | -0.63437  | -0.12064 | 0.087295 |
    +--------------+---------+-----------+-----------+----------+----------+
    | Kurtosis     | 0.34012 | -0.032159 | 0.37947   | -0.16908 | -0.18817 |
    +--------------+---------+-----------+-----------+----------+----------+
    +--------------+---------+-----------+-----------+----------+----------+
    | Minimum      | 147.03  | 19.234    | 11.639    | 154.89   | -0.37247 |
    +--------------+---------+-----------+-----------+----------+----------+
    | 1st Quartile | 149.22  | 19.833    | 13.642    | 158.87   | 2.9019   |
    +--------------+---------+-----------+-----------+----------+----------+
    | Median       | 149.97  | 20.012    | 14.033    | 160.02   | 3.9264   |
    +--------------+---------+-----------+-----------+----------+----------+
    | 3rd Quartile | 150.56  | 20.173    | 14.481    | 161.35   | 5.2160   |
    +--------------+---------+-----------+-----------+----------+----------+
    | Maximum      | 153.27  | 20.505    | 15.460    | 164.51   | 8.2828   |
    +--------------+---------+-----------+-----------+----------+----------+
    +--------------+---------+-----------+-----------+----------+----------+
    | N Outliers   | 5       | 1         | 4         | 1        | 0        |
    +--------------+---------+-----------+-----------+----------+----------+

    """
    data: pd.DataFrame = field(repr=False)
    samples: dict[str, Sample] = field(repr=False)
    capabilities: dict[str, Capability] = field(repr=False)

    def __init__(self, data, specs=None, conf=0.95, ddof=1):
        if isinstance(data, pd.Series):
            self.data = data.to_frame()
        elif isinstance(data, pd.DataFrame):
            self.data = data
        else:
            raise ValueError('`data` must be a Series or a DataFrame.')

        self.samples = {
            name: Sample(col, conf=conf, ddof=ddof)
            for name, col
            in self.data.items()
        }

        if specs is not None:
            self.capabilities = {
                name: Capability(self.samples[name], spec)
                for name, spec
                in specs.items()
            }
        else:
            self.capabilities = {}

    def __getitem__(self, index):
        return self.samples[index]



import mqr

import numpy as np
import pandas as pd
import pytest
import scipy

def test_Summary():
    np.random.seed(0)
    data = pd.DataFrame({
        'x': scipy.stats.norm(1, 2).rvs(100)
    })
    summary = mqr.process.Summary(data)

def test_Capability():
    np.random.seed(0)
    data = pd.DataFrame({
        'x': scipy.stats.norm(1, 2).rvs(100)
    })
    spec = mqr.process.Specification(-1, -1-5*2, -1+5*2) # Potential around 1.67, capability around 1.33
    summary = mqr.process.Summary(data, {'x': spec})

    c = summary.capabilities['x']
    assert c.cp == pytest.approx(1.67, abs=0.1)
    assert c.cpk == pytest.approx(1.33, abs=0.1)



"""
============================================
Measurement system analysis (:mod:`mqr.msa`)
============================================

.. currentmodule:: mqr.msa

User guide
    :doc:`/user_guide/measurement-system-analysis`

Detailed examples
    https://github.com/nklsxn/mqrpy-guide

Construction and presentation of gauge repeatability and reproducibility study.

.. rubric:: Construction
.. autosummary::
    :toctree: generated/

    GRR
    NameMapping

.. rubric:: Results
.. autosummary::
    :toctree: generated/

    VarianceTable
"""

from dataclasses import dataclass, field
import numpy as np
import pandas as pd
from patsy import ModelDesc
import statsmodels
from statsmodels.formula.api import mixedlm

import mqr

@dataclass
class NameMapping:
    """
    A definition of terms that maps variables in an experiment to standard terms
    in a GRR study.

    Attributes
    ----------
    measurement : str, optional
        Column name referring to the observation/KPI. Default 'measurement'.
    part : str, optional
        Column name referring to the categorical part ID. Default 'part'.
    operator : str, optional
        Column name referring to the categorical operator ID. Default 'operator'.
    replicate : str, optional
        Column name referring to the categorical replicate ID. Default 'replicate'.
    """
    measurement: str = field(default='measurement')
    part: str = field(default='part')
    operator: str = field(default='operator')
    _m: str = field(repr=False, default=None)
    _p: str = field(repr=False, default=None)
    _o: str = field(repr=False, default=None)

    def __init__(self, *, measurement=None, part=None, operator=None):
        """
        Construct NameMapping.

        Parameters
        ----------
        measurement : str, optional
            see Attribute.
        part : str, optional
            see Attribute.
        operation : str, optional
            see Attribute.
        """
        if measurement:
            self.measurement = self._m = measurement
        if part:
            self.part = self._p = part
        if operator:
            self.operator = self._o = operator

@dataclass
class GRR:
    """
    A Gauge Repeatability and Reliability study.

    Constructs a model and calls `statsmodels.formula.api.ols`.

    Attributes
    ----------
    data : pd.DataFrame
        Experiment runs and measurements, including columns for measurement, part,
        operator and replicate. See `NameMapping` for how to name these columns.
    tolerance : np.float64
        Width of the tolerance of the process in the same units as the measurements.
    names : NameMapping
        A name mapping that defines how custom names translate to the standard
        names used in this library. See `mqr.msa.NameMapping`.
    include_interaction : bool
        When `True`, include terms in the ANOVA for the interaction between
        operator and part.
    include_intercept : bool
        When `True`, include terms in the ANOVA for the intercept.
    nsigma : int
        Target capability of the process.
    formula : str, automatic
        Formula passed to statsmodels for regression.
    counts : tuple[int], automatic
        Number of measurements, and number of unique levels in categorical
        variables part, operator and replicate.
    model : automatic
        Linear model for OLS.
    regression_result : statsmodels.regression.linear_model.RegressionResultsWrapper, automatic
        Result of calling `fit()` on the `model`.
    """
    data: pd.DataFrame
    tolerance: np.float64
    names: NameMapping
    include_interaction: bool
    include_intercept: bool
    nsigma: int

    formula: str
    counts: tuple[int]

    model: statsmodels.regression.linear_model.OLS
    regression_result: statsmodels.regression.linear_model.RegressionResultsWrapper

    def __init__(
        self,
        data:pd.DataFrame,
        tolerance:float,
        names:NameMapping=None,
        include_interaction=True,
        include_intercept=True,
        nsigma=6):
        """
        Construct GRR.

        Arguments
        ---------
        data : pd.DataFrame
            See attribute.
        tolerance : float
            See attribute.
        names : mqr.msa.NameMapping, optional
            See attribute. Defaults to `NameMapping()`.
        include_interaction : bool, optional
            See attribute.
        include_intercept : bool, optional
            See attribute.
        nsigma : float
            See attribute.
        """
        self.data = data
        self.tolerance = tolerance
        self.include_interaction = include_interaction
        self.include_intercept = include_intercept
        self.names = names if names is not None else NameMapping()
        self.nsigma = nsigma

        self._configure_counts()
        self._configure_formula()

        self._fit_model(data)

    def _configure_counts(self):
        cols = [self.names._p, self.names._o]
        data = self.data.loc[:, cols]
        self.counts = data.nunique(axis=0)
        reps = len(data) / np.prod(self.counts)
        if reps != int(reps):
            raise ValueError('Unbalanced designs not supported.')
        self.replicates = int(reps)

    def _configure_formula(self):
        name_m = self.names._m
        name_p = self.names._p
        name_o = self.names._o
        combn = '*' if self.include_interaction else '+'
        intercept = '+ 1' if self.include_intercept else '- 1'
        formula = f'{name_m} ~ C({name_p}) {combn} C({name_o}) {intercept}'
        self.formula = ModelDesc.from_formula(formula).describe()

    def _fit_model(self, data):
        self.model = statsmodels.formula.api.ols(self.formula, self.data)
        self.regression_result = self.model.fit()

    def _repr_html_(self):
        return SummaryTable(self)._repr_html_()

class SummaryTable:
    _grr: GRR

    def __init__(self, grr: GRR):
        self._grr = grr

    def _repr_html_(self):
        grr = self._grr
        html = f'''
            <table>
            <thead>
                <caption>Gauge Repeatability and Reliability Study</caption>
            </thead>
            <tbody>
                <thead>
                    <tr>
                        <th scope='col'></th>
                        <th scope='col'>Measurement</td>
                        <th scope='col'>Part</td>
                        <th scope='col'>Operator</td>
                    </tr>
                </thead>
                <tr>
                    <th scope='row'>Variable</th>
                    <td>{grr.names.measurement}</td>
                    <td>{grr.names.part}</td>
                    <td>{grr.names.operator}</td>
                </tr>
                <tr>
                    <th scope='row'>Count</th>
                    <td>{grr.data.shape[0]}</td>
                    <td>{grr.counts[grr.names.part]}</td>
                    <td>{grr.counts[grr.names.operator]}</td>
                </tr>
                <thead><tr></tr></thead>
                <tr>
                    <th scope='row'>Replicates</th>
                    <td>{grr.replicates}</td>
                    <td colspan='2'></td>
                </tr>
                <tr>
                    <th scope='row'>Tolerance</th>
                    <td>{grr.tolerance}</td>
                    <td colspan='2'></td>
                </tr>
                <tr>
                    <th scope='row'>N<sub>&#x03C3;</sub></th>
                    <td>{grr.nsigma}</td>
                    <td colspan='2'></td>
                </tr>
                <thead><tr></tr></thead>
                <tr>
                    <th scope='row'>Formula</th>
                    <td colspan='3'>{grr.formula}</td>
                </tr>
            </tbody>
            </table>
            '''
        return html

class VarianceTable:
    """
    GRR variance components

    Variance components are calculated using the method in [1]_.

    Attributes
    ----------
    grr : :class:`GRR`
        GRR study object used to create the variance components.
    anova_table : pandas.DataFrame, automatic
        ANOVA summary table.
    table : pandas.DataFrame
        Table of variance components from GRR model.
    num_distinct_cat : float
        Number of confidence intervals on part measurements that will span the
        part variability (without overlap). For example, if `num_distinct_cat`
        is 5 then the measurement system can discern five groups of parts.
    discrimination : float
        Discrimination ratio, see [1]_.

    References
    ----------
    .. [1]  Montgomery, D. C. (2009).
            Statistical quality control (Vol. 7).
            New York: Wiley.
    """
    grr: NameMapping
    anova_table: pd.DataFrame
    table: pd.DataFrame
    num_distinct_cats: np.float64
    discrimination: np.float64

    def _table_index(self):
        return [
            'Gauge RR',
            'Repeatability',
            'Reproducibility',
            'Operator',
            'Operator*Part',
            'Part-to-Part',
            'Total']

    def _table_columns(self):
        return [
            'VarComp',
            '% Contribution',
            'StdDev',
            f'StudyVar ({self.grr.nsigma}*SD)',
            '% StudyVar',
            '% Tolerance']

    def _table_styles(self):
        if self.grr.include_interaction:
            reprod_indent = 'th.row3,th.row4'
        else:
            reprod_indent = 'th.row3'
        return [
            {
                'selector': '.row_heading',
                'props': [
                    ('text-align', 'left'),
                ]
            },
            {
                'selector': 'th.row1,th.row2',
                'props': [
                    ('padding-left', '1.5em'),
                ]
            },
            {
                'selector': reprod_indent,
                'props': [
                    ('padding-left', '3em'),
                ]
            }
        ]

    def __init__(self, grr: GRR, typ=2):
        self.grr = grr
        self.anova_table = mqr.anova.summary(
            grr.regression_result,
            typ,
            formatted=False)
        self._varcomp()
        self._calculate_table()
        self._set_discrimination()
        self._set_num_distinct_cats()

    def _varcomp(self):
        name_p = self.grr.names._p
        name_o = self.grr.names._o

        N_r = self.grr.replicates
        N_p = self.grr.counts[name_p]
        N_o = self.grr.counts[name_o]

        anova_table = self.anova_table
        MS_e = anova_table.iloc[-2, 2]
        MS_p = anova_table.iloc[0, 2]
        MS_o = anova_table.iloc[1, 2]
        if self.grr.include_interaction:
            MS_i = anova_table.iloc[2, 2]
        else:
            MS_i = 0

        var = MS_e
        var_i = np.clip((MS_i - MS_e) / N_r, 0, np.inf)
        var_p = (MS_p - MS_i) / (N_r * N_o)
        var_o = (MS_o - MS_i) / (N_r * N_p)

        self._variance_components = var_p, var_o, var_i, var

    def _calculate_table(self):
        var_p, var_o, var_i, var = self._variance_components

        table = pd.DataFrame(
            index=self._table_index(),
            columns=self._table_columns(),
            dtype=np.float64)
        table.iloc[:, 0] = [
            var_o + var_i + var,         # GRR
                var,                     # Repeatability
                var_o + var_i,           # Reproducibility
                    var_o,               # Operator
                    var_i,               # Interaction
            var_p,                       # Part-to-Part
            var_p + var_o + var_i + var, # Total
        ]
        table.iloc[:, 1] = 100 * table.iloc[:, 0] / table.iloc[-1, 0]
        table.iloc[:, 2] = np.sqrt(table.iloc[:, 0])
        table.iloc[:, 3] = self.grr.nsigma * table.iloc[:, 2]
        table.iloc[:, 4] = 100 * table.iloc[:, 3] / table.iloc[-1, 3]
        table.iloc[:, 5] = 100 * table.iloc[:, 3] / self.grr.tolerance

        if not self.grr.include_interaction:
            table.drop(index=['Operator*Part'], inplace=True)

        self.table = table

    def _set_discrimination(self):
        rho_p = self.table.loc['Part-to-Part', '% Contribution'] / 100
        self.discrimination = (1 + rho_p) / (1 - rho_p)

    def _set_num_distinct_cats(self):
        rho_p = self.table.loc['Part-to-Part', '% Contribution'] / 100
        self.num_distinct_cats = np.sqrt(2 * rho_p / (1 - rho_p))

    def _repr_html_(self):
        n_cats = int(np.floor(self.num_distinct_cats))
        html = '<div style="display:flex; flex-direction:column; align-items:flex-start;">'
        html += (
            self.table.style
            .format(formatter='{:.4g}')
            .set_table_styles(self._table_styles())
            ._repr_html_())
        html += f'<div><b>Number of distinct categories:</b> {n_cats:d}</div>'
        html += '</div>'
        return html

# class ConfTable:
#     def _make_conf_int_table(self):
#         p, o, n = self.counts
#         [MS_p, MS_o, MS_i, MS_e] = self._mean_squares
#         var_grr = self.grr_table.loc['Gauge RR', 'VarComp']
#         var_p = self.grr_table.loc['Part-to-Part', 'VarComp']
#         var_tot = self.grr_table.loc['Total', 'VarComp']
#         intervals = _conf_int(p, o, n, MS_p, MS_o, MS_i, MS_e, self.alpha)
#         [
#             (var_p_lower, var_p_upper),
#             (var_grr_lower, var_grr_upper),
#             (var_tot_lower, var_tot_upper),
#             (rho_p_lower, rho_p_upper),
#             (rho_m_lower, rho_m_upper),
#         ] = intervals

#         values = np.array([
#             [var_grr, var_grr + var_grr_lower, var_grr + var_grr_upper],
#             [var_p, var_p + var_p_lower, var_p + var_p_upper],
#             [var_tot, var_tot + var_tot_lower, var_tot + var_tot_upper],
#             [np.nan, rho_m_lower*100, rho_m_upper*100],
#             [np.nan, rho_p_lower*100, rho_p_upper*100],
#         ])
#         table = pd.DataFrame(
#             values,
#             index=GRR._conf_index(),
#             columns=GRR._conf_columns(self.alpha))
#         self.conf_int_table = table

    # @staticmethod
    # def _conf_index():
    #     return [
    #         'Gauge RR',
    #         'Part-to-Part',
    #         'Total',
    #         'Gauge RR (%)',
    #         'Part-to-Part (%)',
    #     ]

    # @staticmethod
    # def _conf_columns(alpha):
    #     return [
    #         f'E(var)',
    #         f'[{100*alpha/2:.3g}%',
    #         f'{100*(1-alpha/2):.3g}%]'
    #     ]


# def _conf_int(p, o, n, MS_P, MS_O, MS_PO, MS_E, alpha=0.05):
#     # Burdick, Richard K., Connie M. Borror, and Douglas C. Montgomery. "A review of methods for measurement systems capability analysis." Journal of Quality Technology 35.4 (2003): 342-354.

#     V_LP, V_UP, V_LM, V_UM, V_LT, V_UT, L_star, U_star = _conf_int_intermediate_vals(p, o, n, MS_P, MS_O, MS_PO, MS_E, alpha)
    
#     var_p_lower = -np.sqrt(V_LP) / (o * n)
#     var_p_upper = np.sqrt(V_UP) / (o * n)

#     var_grr_lower = -np.sqrt(V_LM) / (p * n)
#     var_grr_upper = np.sqrt(V_UM) / (p * n)

#     var_tot_lower = -np.sqrt(V_LT) / (p * o * n)
#     var_tot_upper = np.sqrt(V_UT) / (p * o * n)

#     rho_p_lower = L_p = p * L_star / (p * L_star + o)
#     rho_p_upper = U_p = p * U_star / (p * U_star + o)

#     rho_m_lower = 1 - U_p
#     rho_m_upper = 1 - L_p

#     return [
#         (var_p_lower, var_p_upper),
#         (var_grr_lower, var_grr_upper),
#         (var_tot_lower, var_tot_upper),
#         (rho_p_lower, rho_p_upper),
#         (rho_m_lower, rho_m_upper),
#     ]

# def _conf_int_intermediate_vals(p, o, n, MS_P, MS_O, MS_PO, MS_E, alpha=0.05):
#     # Burdick, Richard K., Connie M. Borror, and Douglas C. Montgomery. "A review of methods for measurement systems capability analysis." Journal of Quality Technology 35.4 (2003): 342-354.
#     # But doesn't match "Introduction to Statistical Quality Control", Montgomery... need to look a bit closer at this.

#     # Distributions
#     F_p = st.chi2(p-1)
#     F_o = st.chi2(o-1)
#     F_i = st.chi2((p-1)*(o-1))
#     F_e = st.chi2(p*o*(n-1))
#     F_po = st.f(p-1, o-1)
#     F_pi = st.f(p-1, (p-1)*(o-1))

#     # Intermediate quantities
#     G_1 = 1 - 1 / F_p.ppf(1-alpha/2)
#     G_2 = 1 - 1 / F_o.ppf(1-alpha/2)
#     G_3 = 1 - 1 / F_i.ppf(1-alpha/2)
#     G_4 = 1 - 1 / F_e.ppf(1-alpha/2)

#     H_1 = 1 / F_p.ppf(alpha/2) - 1
#     H_2 = 1 / F_o.ppf(alpha/2) - 1
#     H_3 = 1 / F_i.ppf(alpha/2) - 1
#     H_4 = 1 / F_e.ppf(alpha/2) - 1

#     G_13 = ((F_pi.ppf(1-alpha/2)-1)**2 - G_1**2 * F_pi.ppf(1-alpha/2)**2 - H_3**2) / F_pi.ppf(1-alpha/2)
#     H_13 = ((1-F_pi.ppf(alpha/2))**2 - H_1**2 * F_pi.ppf(alpha/2)**2 - G_3**2) / F_pi.ppf(alpha/2)

#     V_LP = G_1**2 * MS_P**2 + H_3**2 * MS_PO**2 + G_13 * MS_P
#     V_UP = H_1**2 * MS_P**2 + G_3**2 * MS_PO**2 + H_13 * MS_P**2 * MS_PO

#     V_LM = G_2**2 * MS_O**2 + G_3**2 * (p-1)**2 * MS_PO**2 + G_4**2 * p**2 * (n-1)**2 * MS_E**2
#     V_UM = H_2**2 * MS_O**2 + H_3**2 * (p-1)**2 * MS_PO**2 + H_4**2 * p**2 * (n-1)**2 * MS_E**2

#     V_LT = G_1**2 * p**2 * MS_P**2 + G_2**2 * o**2 * MS_O**2 + G_3**2 * (p*o - p - o)**2 * MS_PO**2 + G_4**2 * (p*o)**2 * (n-1)**2 * MS_E**2
#     V_UT = H_1**2 * p**2 * MS_P**2 + H_2**2 * o**2 * MS_O**2 + H_3**2 * (p*o - p - o)**2 * MS_PO**2 + H_4**2 * (p*o)**2 * (n-1)**2 * MS_E**2

#     L_star_num = MS_P - F_pi.ppf(1-alpha/2) * MS_PO
#     L_star_den = p * (n-1) * F_p.ppf(1-alpha/2) * MS_E + F_po.ppf(1-alpha/2) * MS_O + (p-1) * F_p.ppf(1-alpha/2) * MS_PO
#     L_star = L_star_num / L_star_den

#     U_star_num = MS_P - F_pi.ppf(alpha/2) * MS_PO
#     U_star_den = p * (n-1) * F_p.ppf(alpha/2) * MS_E + F_po.ppf(alpha/2) * MS_O + (p-1) * F_p.ppf(alpha/2) * MS_PO
#     U_star = U_star_num / U_star_den

#     return V_LP, V_UP, V_LM, V_UM, V_LT, V_UT, L_star, U_star



from dataclasses import dataclass
import numbers
import numpy as np

@dataclass
class TestPower:
    """
    Result of a sample size calculation for an hypothesis test.

    Attributes
    ----------
    name : str
        Description of the hypothesis test.
    alpha : float
        Required significance level.
    beta : float
        Complement of the required power: 1 - power.
    effect : float
        Required effect size of the test.
    alternative : {'two-sided', 'less', 'greater'}
        Sense of the test alternative.
    method : str
        Name of the hypothesis test method, if applicable. For example, 't-test'.
    sample_size : int
        Lower bound on the sample size to achieve the above parameters.

    Examples
    --------
    Calculations solving for power or sample size produce types like this.

    >>> mqr.inference.power.TestPower(
    >>>     name="PowerName",
    >>>     alpha=0.01,
    >>>     beta=0.10,
    >>>     effect=1.5,
    >>>     alternative="greater",
    >>>     method="TestMethod",
    >>>     sample_size=45)
    TestPower(
        name='PowerName',
        alpha=0.05, beta=0.1, effect=1.5,
        alternative='greater', method='TestMethod',
        sample_size=45)

    In jupyter notebooks, power calculations are rendered as an HTML table:

    +-----------------+------------+
    | | **Test Power**             |
    | | PowerName                  |
    +=================+============+
    | **alpha**       | 0.05       |
    +-----------------+------------+
    | **beta**        | 0.1        |
    +-----------------+------------+
    | **effect**      | 1.5        |
    +-----------------+------------+
    | **alternative** | greater    |
    +-----------------+------------+
    | **method**      | TestMethod |
    +-----------------+------------+
    | **sample size** | 45         |
    +-----------------+------------+

    """
    name: str
    alpha: np.float64
    beta: np.float64
    effect: np.float64
    alternative: str
    method: str
    sample_size: int

    def _html(self):
        if isinstance(self.effect, numbers.Number):
            effect_str = f'{self.effect:g}'
        else:
            effect_str = str(self.effect)

        return f'''
        <table>
        <thead>
            <tr>
                <th scope="col" colspan=2 style="text-align: left; padding-bottom: 0px;">Test Power</th>
            </tr>
            <tr style='padding-top: 0px;'>
                <td colspan=2 style='text-align: left; padding-top: 0px'>{self.name}</td>
            </tr>
        </thead>
        <tbody>
            <tr>
                <th scope='row' style='text-align: left;'>alpha</td>
                <td style='text-align: left;'>{self.alpha:g}</td>
            </tr>
            <tr>
                <th scope='row' style='text-align: left;'>beta</td>
                <td style='text-align: left;'>{self.beta:g}</td>
            </tr>
            <tr>
                <th scope='row' style='text-align: left;'>effect</td>
                <td style='text-align: left;'>{effect_str}</td>
            </tr>
            <tr>
                <th scope='row' style='text-align: left;'>alternative</td>
                <td style='text-align: left;'>{self.alternative}</td>
            </tr>
            <tr>
                <th scope='row' style='text-align: left;'>method</td>
                <td style='text-align: left;'>{self.method}</td>
            </tr>
            <thead><tr/></thead>
            <tr>
                <th scope='row' style='text-align: left;'>sample size</td>
                <td style='text-align: left;'>{self.sample_size:g}</td>
            </tr>
        </tbody>
        </table>
        '''

    def _repr_html_(self):
        return self._html()



import numbers
import numpy as np
import pytest

import mqr

def test_size_1sample():
    alpha = 0.10
    beta = 0.10

    effect = np.sqrt(1.5)
    alternative = 'two-sided'
    res = mqr.inference.stddev.size_1sample(effect, alpha, beta, alternative)
    assert res.name == 'standard deviation'
    assert res.alpha == alpha
    assert res.beta == beta
    assert res.effect == effect
    assert res.alternative == alternative
    assert res.method == 'chi2'
    assert np.ceil(res.sample_size) == 105

    alternative = 'greater'
    res = mqr.inference.stddev.size_1sample(effect, alpha, beta, alternative)
    assert res.name == 'standard deviation'
    assert res.alpha == alpha
    assert res.beta == beta
    assert res.effect == effect
    assert res.alternative == alternative
    assert res.method == 'chi2'
    assert np.ceil(res.sample_size) == 82

    effect = np.sqrt(0.8)
    alternative = 'less'
    res = mqr.inference.stddev.size_1sample(effect, alpha, beta, alternative)
    assert res.name == 'standard deviation'
    assert res.alpha == alpha
    assert res.beta == beta
    assert res.effect == effect
    assert res.alternative == alternative
    assert res.method == 'chi2'
    assert np.ceil(res.sample_size) == 266

def test_size_2sample():
    alpha = 0.10
    beta = 0.10

    std_ratio = np.sqrt(1.5)
    alternative = 'two-sided'
    res = mqr.inference.stddev.size_2sample(std_ratio, alpha, beta, alternative)
    assert res.name == 'ratio of standard deviations'
    assert res.alpha == alpha
    assert res.beta == beta
    assert res.effect == std_ratio
    assert res.alternative == alternative
    assert res.method == 'f'
    assert np.ceil(res.sample_size) == 211

    alternative = 'greater'
    res = mqr.inference.stddev.size_2sample(std_ratio, alpha, beta, alternative)
    assert res.name == 'ratio of standard deviations'
    assert res.alpha == alpha
    assert res.beta == beta
    assert res.effect == std_ratio
    assert res.alternative == alternative
    assert res.method == 'f'
    assert np.ceil(res.sample_size) == 162

    std_ratio = np.sqrt(0.8)
    alternative = 'less'
    res = mqr.inference.stddev.size_2sample(std_ratio, alpha, beta, alternative)
    assert res.name == 'ratio of standard deviations'
    assert res.alpha == alpha
    assert res.beta == beta
    assert res.effect == std_ratio
    assert res.alternative == alternative
    assert res.method == 'f'
    assert np.ceil(res.sample_size) == 530

def test_confint_1sample():
    x = np.array([0, 1, 2])
    conf = 0.90

    bounded = 'both'
    res = mqr.inference.stddev.confint_1sample(x, conf, bounded=bounded)
    assert res.name == 'standard deviation'
    assert res.method == 'chi2'
    assert res.value == np.std(x, ddof=1)
    assert res.lower == pytest.approx(np.sqrt(0.3338), 1e-4)
    assert res.upper == pytest.approx(np.sqrt(19.4957), 1e-4)
    assert res.conf == conf
    assert res.bounded == bounded

    bounded = 'below'
    res = mqr.inference.stddev.confint_1sample(x, conf, bounded=bounded)
    assert res.name == 'standard deviation'
    assert res.method == 'chi2'
    assert res.value == np.std(x, ddof=1)
    assert res.lower == pytest.approx(np.sqrt(0.4343), 1e-4)
    assert res.upper == np.inf
    assert res.conf == conf
    assert res.bounded == bounded

    bounded = 'above'
    res = mqr.inference.stddev.confint_1sample(x, conf, bounded=bounded)
    assert res.name == 'standard deviation'
    assert res.method == 'chi2'
    assert res.value == np.std(x, ddof=1)
    assert res.lower == 0.0
    assert res.upper == pytest.approx(np.sqrt(9.4912), 1e-4)
    assert res.conf == conf
    assert res.bounded == bounded

def test_confint_2sample():
    x = np.array([0, 1, 2])
    y = np.array([0, 2, 4])
    conf = 0.90

    bounded = 'both'
    res = mqr.inference.stddev.confint_2sample(x, y, conf, bounded=bounded)
    assert res.name == 'ratio of standard deviations'
    assert res.method == 'f'
    assert res.value == np.std(x, ddof=1) / np.std(y, ddof=1)
    assert res.lower == pytest.approx(np.sqrt(0.01316), abs=1e-4)
    assert res.upper == pytest.approx(np.sqrt(4.75))
    assert res.conf == conf
    assert res.bounded == bounded

    bounded = 'below'
    res = mqr.inference.stddev.confint_2sample(x, y, conf, bounded=bounded)
    assert res.name == 'ratio of standard deviations'
    assert res.method == 'f'
    assert res.value == np.std(x, ddof=1) / np.std(y, ddof=1)
    assert res.lower == pytest.approx(np.sqrt(0.02778), abs=1e-4)
    assert res.upper == np.inf
    assert res.conf == conf
    assert res.bounded == bounded

    bounded = 'above'
    res = mqr.inference.stddev.confint_2sample(x, y, conf, bounded=bounded)
    assert res.name == 'ratio of standard deviations'
    assert res.method == 'f'
    assert res.value == np.std(x, ddof=1) / np.std(y, ddof=1)
    assert res.lower == 0.0
    assert res.upper == pytest.approx(np.sqrt(2.25))
    assert res.conf == conf
    assert res.bounded == bounded

def test_test_1sample():
    x = np.array([0, 1, 2])
    H0_std = 1

    alternative = 'two-sided'
    res = mqr.inference.stddev.test_1sample(x, H0_std, alternative)
    assert res.description == 'standard deviation'
    assert res.alternative == alternative
    assert res.method == 'chi2'
    assert res.sample_stat == 'std(x)'
    assert res.sample_stat_target == H0_std
    assert res.sample_stat_value == np.std(x, ddof=1)
    assert isinstance(res.stat, numbers.Number)
    assert res.pvalue == pytest.approx(0.7358, abs=1e-4)

    alternative = 'greater'
    res = mqr.inference.stddev.test_1sample(x, H0_std, alternative)
    assert res.description == 'standard deviation'
    assert res.alternative == alternative
    assert res.method == 'chi2'
    assert res.sample_stat == 'std(x)'
    assert res.sample_stat_target == H0_std
    assert res.sample_stat_value == np.std(x, ddof=1)
    assert isinstance(res.stat, numbers.Number)
    assert res.pvalue == pytest.approx(0.3679, abs=1e-4)

    alternative = 'less'
    res = mqr.inference.stddev.test_1sample(x, H0_std, alternative)
    assert res.description == 'standard deviation'
    assert res.alternative == alternative
    assert res.method == 'chi2'
    assert res.sample_stat == 'std(x)'
    assert res.sample_stat_target == H0_std
    assert res.sample_stat_value == np.std(x, ddof=1)
    assert isinstance(res.stat, numbers.Number)
    assert res.pvalue == pytest.approx(0.6321, abs=1e-4)

def test_test_2sample():
    x = np.array([0, 1, 2])
    y = np.array([1, 2, 4])

    method = 'f'
    alternative = 'two-sided'
    res = mqr.inference.stddev.test_2sample(x, y, alternative)
    assert res.description == 'ratio of standard deviations'
    assert res.alternative == alternative
    assert res.method == method
    assert res.sample_stat == 'std(x) / std(y)'
    assert res.sample_stat_target == 1.0
    assert res.sample_stat_value == np.std(x, ddof=1) / np.std(y, ddof=1)
    assert isinstance(res.stat, numbers.Number)
    assert res.pvalue == pytest.approx(0.6)

    alternative = 'less'
    res = mqr.inference.stddev.test_2sample(x, y, alternative)
    assert res.description == 'ratio of standard deviations'
    assert res.alternative == alternative
    assert res.method == method
    assert res.sample_stat == 'std(x) / std(y)'
    assert res.sample_stat_target == 1.0
    assert res.sample_stat_value == np.std(x, ddof=1) / np.std(y, ddof=1)
    assert isinstance(res.stat, numbers.Number)
    assert res.pvalue == pytest.approx(0.3)

    alternative = 'greater'
    res = mqr.inference.stddev.test_2sample(x, y, alternative)
    assert res.description == 'ratio of standard deviations'
    assert res.alternative == alternative
    assert res.method == method
    assert res.sample_stat == 'std(x) / std(y)'
    assert res.sample_stat_target == 1.0
    assert res.sample_stat_value == np.std(x, ddof=1) / np.std(y, ddof=1)
    assert isinstance(res.stat, numbers.Number)
    assert res.pvalue == pytest.approx(0.7)



import importlib.metadata
__version__ = importlib.metadata.version('mqrpy')

import mqr.anova
import mqr.doe
import mqr.inference
import mqr.msa
import mqr.nbtools
import mqr.plot
import mqr.process
import mqr.spc
import mqr.transforms

# Sample data
def sample_data(name):
    import importlib
    return importlib.resources.files('mqr.data')/name

# Notebook formatters
import mqr.notebook



'''
Check call-throughs.
'''

import mqr.utils as utils

import numpy as np
import pytest

def test_make_ordinal():
    assert utils.make_ordinal(1) == '1st'
    assert utils.make_ordinal(2) == '2nd'
    assert utils.make_ordinal(3) == '3rd'
    assert utils.make_ordinal(4) == '4th'
    assert utils.make_ordinal(5) == '5th'

    assert utils.make_ordinal(10) == '10th'
    assert utils.make_ordinal(11) == '11th'
    assert utils.make_ordinal(12) == '12th'
    assert utils.make_ordinal(13) == '13th'
    assert utils.make_ordinal(14) == '14th'

    assert utils.make_ordinal(21) == '21st'
    assert utils.make_ordinal(22) == '22nd'
    assert utils.make_ordinal(23) == '23rd'
    assert utils.make_ordinal(24) == '24th'
    assert utils.make_ordinal(25) == '25th'

    assert utils.make_ordinal(1.1) == '1.1th'
    assert utils.make_ordinal(5.1) == '5.1th'
    assert utils.make_ordinal(134.1) == '134.1th'

def test_clip_where():
    assert utils.clip_where(0, -1, 1, True) == 0
    assert utils.clip_where(0, 1, 2, True) == 1
    assert utils.clip_where(0, -2, -1, True) == -1

    assert utils.clip_where(0, -1, 1, False) == 0
    assert utils.clip_where(0, 1, 2, False) == 0
    assert utils.clip_where(0, -2, -1, False) == 0

    a = np.array([1, 1, 1, 1, 1, 1])
    res = utils.clip_where(
        a,
        np.array([0, 0, 1, 1, 2, 2]),
        np.inf,
        np.array([1, 0, 1, 0, 1, 0]))
    assert list(res) == [1, 1, 1, 1, 2, 1]
    res = utils.clip_where(
        a,
        -np.inf,
        np.array([0, 0, 1, 1, 2, 2]),
        np.array([1, 0, 1, 0, 1, 0]))
    assert list(res) == [0, 1, 1, 1, 1, 1]
    res = utils.clip_where(
        a,
        np.array([0, 0, 1, 1, 2, 2]),
        np.array([0, 0, 1, 1, 2, 2]),
        np.array([1, 0, 1, 0, 1, 0]))
    assert list(res) == [0, 1, 1, 1, 2, 1]

    a = np.array([1, 1, 1, 1, 1, 1])
    res = utils.clip_where(
        a,
        2,
        3,
        np.array([1, 0, 1, 0, 1, 0]))
    assert list(res) == [2, 1, 2, 1, 2, 1]
    res = utils.clip_where(
        a,
        np.array([0, 0, 1, 1, 2, 2]),
        np.array([0, 0, 1, 1, 2, 2]),
        True)
    assert list(res) == [0, 0, 1, 1, 2, 2]
    res = utils.clip_where(a, 2, 3, True)
    list(res) == [0, 0, 1, 1, 2, 2]
    res = utils.clip_where(a, 2, 3, False)
    list(res) == [1, 1, 1, 1, 1, 1]

def test_fredholm2_1():
    fn_g = lambda t: np.sin(t)
    lmda = 1
    fn_K = lambda t, s: np.sin(t) * np.cos(s)

    c, d = 0, np.pi / 2
    x, w = np.polynomial.legendre.leggauss(10)
    x = c + (x + 1) * (d - c) / 2
    w = w * (d - c) / 2

    t0 = np.linspace(c, d)
    exact_result = 2 * np.sin(t0)

    for i in range(len(t0)):
        fred_result = utils.fredholm2(
            t0=t0[i],
            fn_K=fn_K,
            fn_g=fn_g,
            lmda=lmda,
            x=x,
            w=w)
        assert exact_result[i] == pytest.approx(fred_result, abs=1e-15)

def test_fredholm2_2():
    fn_g = lambda t: 1
    lmda = 1
    fn_K = lambda t, s: t

    alpha = 0.9876
    c, d = 0, alpha
    x, w = np.polynomial.legendre.leggauss(10)
    x = c + (x + 1) * (d - c) / 2
    w = w * (d - c) / 2

    t0 = np.linspace(c, d)
    exact_result = 1 + 2 * alpha / (2 - alpha**2) * t0

    for i in range(len(t0)):
        fred_result = utils.fredholm2(
            t0=t0[i],
            fn_K=fn_K,
            fn_g=fn_g,
            lmda=lmda,
            x=x,
            w=w)
        assert exact_result[i] == pytest.approx(fred_result, abs=1e-15)



"""
================================================
Regression analysis (:mod:`mqr.plot.regression`)
================================================

.. currentmodule:: mqr.plot.regression

.. autosummary::
    :toctree: generated/

    residuals
    influence
    res_probplot
    res_histogram
    res_v_obs
    res_v_fit
    res_v_factor

Examples
--------
This example uses :func:`residuals` which displays :func:`res_probplot`,
:func:`res_histogram`, :func:`res_v_obs` and :func:`res_v_fit` into four supplied
axes. The example also plots th residuals against factors using :func:`res_v_factor`.
Finally, the example overlay on the `res_v_obs` plot an influence statistic using
:func:`influence`.

First set up the data and fit a model whose residuals will be shown.

.. plot::
    :context: close-figs
    :nofigs:

    import statsmodels

    # Raw data
    data = pd.read_csv(mqr.sample_data('anova-glue.csv'), index_col='Run')

    # Fit a linear model
    model = statsmodels.formula.api.ols('adhesion_force ~ C(primer) * C(glue)', data)
    result = model.fit()

Now create the residuals plots. There are six axes altogether: four for
:func:`residuals` and another two for :func:`res_v_factor` applied to each factor.

.. plot::
    :context: close-figs

    fig, axs = plt.subplots(3, 2, figsize=(8, 5), layout='constrained')

    # show the four residuals plots
    mqr.plot.regression.residuals(result.resid, result.fittedvalues, axs=axs[:2, :])

    # plot residuals against each factor
    mqr.plot.regression.res_v_factor(result.resid, data['primer'], axs[2, 0])
    mqr.plot.regression.res_v_factor(result.resid, data['glue'], axs[2, 1])

    # show Cook's Distance measure of influence.
    mqr.plot.regression.influence(result, 'cooks_dist', axs[1, 0])

"""

import numpy as np
import probscale
import statsmodels.api as sm
import scipy.stats as st
import seaborn as sns

import mqr
from mqr.plot.defaults import Defaults
from mqr.plot.lib.util import set_kws

def res_probplot(resid, ax, probplot_kws=None):
    """
    Probability plot of residuals from result of calling `fit()` on statsmodels
    model.

    Parameters
    ----------
    resid : array_like
        Residuals from regression.
    ax : matplotlib.axes.Axes
        Axes for the plot.
    probplot_kws : dict, optional
        Keyword arguments passed to `probscale.probplot`.
    """
    probplot_kws = set_kws(
        probplot_kws,
        probax='y',
        bestfit=True,
        scatter_kws=dict(
            color='C0',
            marker=Defaults.marker,
            zorder=0,
        ),
        line_kws=dict(
            color='C1',
            alpha=0.8,
            path_effects=Defaults.line_overlay_path_effects,
            zorder=1,
        ),
    )

    probscale.probplot(resid, ax=ax, **probplot_kws)
    ax.set_yticks([1, 5, 20, 50, 80, 95, 99])
    ax.set_ylabel('probability')

def res_histogram(resid, ax, show_density=True, hist_kws=None, density_kws=None):
    """
    Plot histogram of residuals from result of calling `fit()` on statsmodels
    model.

    Parameters
    ----------
    resid : array_like
        Residuals from regression.
    ax : matplotlib.axes.Axes
        Axes for the plot.
    show_density : bool, optional
        Draw a fitted normal distribution density over the histogram.
    hist_kws : dict, optional
        Keyword arguments passed to `seaborn.histplot`.
    density_kws : dict, optional
        Keyword arguments passed to `matplotlib.pyplot.plot`.
    """
    hist_kws = set_kws(
        hist_kws,
        color='C0',
        bins='doane',
        stat='count',
        zorder=0,
    )
    density_kws = set_kws(
        density_kws,
        color='C1',
        path_effects=Defaults.line_overlay_path_effects,
        zorder=1,
    )

    sns.histplot(resid, ax=ax, **hist_kws)
    if show_density:
        mean = np.mean(resid)
        std = np.std(resid, ddof=1)
        xs = np.linspace(mean-3*std, mean+3*std, 200)
        ys = mqr.plot.lib.util.scaled_density(xs, resid, st.norm, hist_kws['bins'])
        ax.plot(xs, ys, **density_kws)
    ax.set_xlabel('residual')
    ax.set_ylabel('frequency')

def res_v_obs(resid, ax, plot_kws=None, bar_kws=None):
    """
    Plot residuals versus observations.

    Parameters
    ----------
    resid : array_like
        Residuals from regression.
    ax : matplotlib.axes.Axes
        Axes for plot.
    plot_kws : dict, optional
        Keyword arguments passed to `matplotlib.pyplot.plot`.
    bar_kws : dict, optional
        Keyword arguments passed to `matplotlib.pyplot.bar`.
    """
    plot_kws = set_kws(
        plot_kws,
        color='C0',
        marker=Defaults.marker,
    )
    bar_kws = set_kws(
        bar_kws,
        alpha=0.5,
        color='C0',
        zorder=0,
    )

    if hasattr(resid, 'index'):
        index = resid.index
    else:
        index = np.arange(len(resid))

    ax.plot(index, resid, **plot_kws)
    ax.grid(axis='y')
    try:
        ax.set_xlabel(index.name)
    except:
        ax.set_xlabel('run')
    ax.set_ylabel('residual')

def res_v_fit(resid, fitted, ax, plot_kws=None):
    """
    Plot residual versus fit.

    Parameters
    ----------
    resid : array_like
        Residuals from regression.
    fitted : array_like
        Fitted values from regression.
    ax : matplotlib.axes.Axes
        Axes for plot.
    plot_kws : dict, optional
        Keyword arguments passed to `matplotlib.pyplot.plot`.
    """
    plot_kws = set_kws(
        plot_kws,
        color='C0',
        linewidth=0,
        marker=Defaults.marker,
    )

    ax.plot(fitted, resid, **plot_kws)
    ax.grid(axis='y')
    ax.set_xlabel('fitted value')
    ax.set_ylabel('residual')

def res_v_factor(resid, factor, ax, factor_ticks=True, factor_name=None, plot_kws=None):
    """
    Plot a factor versus fit.

    Parameters
    ----------
    resid : array_like
        Residuals from regression.
    factor
        Values of a factor (levels) from data.
    ax : matplotlib.axes.Axes
        Axes for plot.
    factor_ticks : bool, optional
        When `True`, uses unique values in `factor` as x-ticks.
    factor_name : str, optional
        Name of the factor to be printed as the x-axis label.
    plot_kws : dict, optional
        Keyword arguments passed to `matplotlib.pyplot.plot`.
    """
    plot_kws = set_kws(
        plot_kws,
        color='C0',
        linewidth=0,
        marker=Defaults.marker,
    )

    if factor_name is None:
        if hasattr(factor, 'name'):
            factor_name = factor.name
        else:
            factor_name = 'factor'
    ax.plot(factor, resid, **plot_kws)
    if factor_ticks:
        ax.set_xticks(factor.unique())
        ax.set_xmargin(0.2)
    ax.set_xlabel(factor_name)
    ax.set_ylabel('residual')
    ax.grid(axis='y')

def residuals(resid, fitted, axs):
    """
    Plot a probability plot of residuals, histogram of residuals, residuals
    versus observation and residuals versus fitted values for the residuals in
    a fitted statsmodels model.

    Parameters
    ----------
    resid
        Residuals from regression.
    fitted
        Fitted values from regression.
    axs : np.ndarray[matplotlib.axes.Axes]
        Array of axes for plot. Must have four elements. Will be flattened
        before use.

    Examples
    --------
    See :mod:`mqr.plot.regression`.
    """
    axs = axs.flatten()
    assert len(axs) == 4 , f'subplots must have 4 axes.'

    res_probplot(resid, ax=axs[0])
    res_histogram(resid, ax=axs[1])
    res_v_obs(resid, ax=axs[2])
    res_v_fit(resid, fitted, ax=axs[3])

def influence(result, influence_stat, ax, bar_kws=None):
    """
    Plot a bar graph of an influence statistic onto a twin axis of `ax`.

    Parameters
    ----------
    result : statsmodels.regression.linear_model.RegressionResults
        Result of calling `fit` on a statsmodel linear regression model.
    influence_stat: {'cooks_dist', 'bonferroni'}
        Plot this influence statistic for each residual as a bar.
    ax : matplotlib.axes.Axes
        Axes from which the twin axes will be created for the plot.
    bar_kws : dict, optional
        Keyword arguments passed to `matplotlib.pyplot.bar`.

    Examples
    --------
    See :mod:`mqr.plot.regression`.
    """
    bar_kws = set_kws(
        bar_kws,
        alpha=0.5,
        color='C0',
        zorder=0,
    )

    if hasattr(result.resid, 'index'):
        index = result.resid.index
    else:
        index = np.arange(len(resid))

    ax.set_ylabel('residual')

    if influence_stat == 'cooks_dist':
        p = 1 - result.get_influence().cooks_distance[1]
        label = "Cook's distance (1-p)"
    elif influence_stat == 'bonferroni':
        p = 1 - result.outlier_test().loc[:, 'bonf(p)']
        label = 'Bonferroni (1-p)'
    else:
        raise RuntimeError(f'statistic not recognised: {influence_stat}')

    axt = ax.twinx()
    axt.bar(index, p, **bar_kws)
    axt.set_ylim(0.0, 1.0)
    axt.set_ylabel(label)



import mqr

import numpy as np
import pandas as pd
import pytest
import scipy
from statsmodels.formula.api import ols

def test_summary():
    """
    Checking properties calculated for the table, but not the properties of the
    regression provided by statsmodels.
    """
    np.random.seed(0)
    data = pd.DataFrame({
        'x': scipy.stats.norm().rvs(20),
        'y': scipy.stats.norm().rvs(20),
        'z': scipy.stats.norm().rvs(20),
    })
    mod = ols('z ~ x + y', data)
    res = mod.fit()

    summary = mqr.anova.summary(res, formatted=False)

    assert (
        summary.loc['Total', 'df'] ==
        summary.loc['Residual', 'df'] + summary.loc['y', 'df'] + summary.loc['y', 'df']
    )
    assert (
        summary.loc['Total', 'sum_sq'] ==
        summary.loc['Residual', 'sum_sq'] + summary.loc['x', 'sum_sq'] + summary.loc['y', 'sum_sq']
    )
    assert list(summary['sum_sq']) == list(summary['mean_sq'] * summary['df'])

def test_groups():
    # Two blocks of zero-mean random noise for the groups
    np.random.seed(0)
    xs = scipy.stats.norm().rvs(20)
    xs[:10] -= np.mean(xs[:10])
    xs[10:] -= np.mean(xs[10:])

    # Measurements, with the second groups mean higher than the first
    ys = xs.copy()
    ys[:10] += 5

    data = pd.DataFrame({
        'a': [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        'b': [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
        'x': ys,
    })

    mod = ols('x ~ C(a) + C(b)', data)
    res = mod.fit()

    groups = mqr.anova.groups(res, value='x', factor='a', conf=0.99, formatted=False)
    assert np.all(np.isclose(
        (groups['upper'] - groups['lower']) / 2,
        scipy.stats.t(18).ppf(0.995) * np.std(xs, ddof=2) / np.sqrt(10)))

def test_interactions():
    data = pd.DataFrame({
        'a': [0, 0, 0, 0, 1, 1, 1, 1],
        'b': [0, 0, 1, 1, 0, 0, 1, 1],
        'c': [0, 1, 0, 1, 0, 1, 0, 1],
        'z': [1.0, 1.2, 1.4, 1.8, 0.75, 0.85, 1.0, 1.6],
    })
    intn = mqr.anova.interactions(data, value='z', between=['a', 'b'], formatted=False)

    assert list(intn.values.flatten()) == [1.1, 1.6, 0.8, 1.3]

def test_adequacy():
    """
    The adequacy table presents data from statsmodels routines.
    Here, checking only that the correct values are passed through.
    """
    data = pd.DataFrame({
        'peas': [1, 1, 0, 0, 0, 0, 0],
        'lupins': [0, 0, 1, 1, 1, 0, 0],
        'mustard': [0, 0, 0, 0, 0, 1, 1],
        'matter': [3.0, 3.6, 3.4, 3.8, 3.3, 4.4, 4.0],
    })
    mod = ols('matter ~ C(peas) + C(lupins) + C(mustard)', data)
    res = mod.fit()
    res = mod.fit()

    adeq = mqr.anova.adequacy(res, formatted=False)
    adeq.loc['', 'R-sq'] == pytest.approx(res.rsquared)
    adeq.loc['', 'R-sq (adj)'] == pytest.approx(res.rsquared_adj)
    adeq.loc['', 'F'] == pytest.approx(res.fvalue)
    adeq.loc['', 'PR(>F)'] == pytest.approx(res.f_pvalue)
    adeq.loc['', 'AIC'] == pytest.approx(res.aic)
    adeq.loc['', 'BIC'] == pytest.approx(res.bic)



from mqr.utils import clip_where
import mqr.inference.lib.util as util

import numpy as np
import scipy

import warnings

def confint_1sample_agresti_coull(count, nobs, conf, bounded):
    """
    Confidence interval for proportion `count / nobs`.

    Implements the Agresti-Coull interval, which is not recommended for small
    sample sizes (small means nobs < 40), according to [1]_. In those cases [1]_
    recommends Wilson with continuity correction or Jeffreys.

    Parameters
    ----------
    count : int
        Number of "true" observations.
    nobs : int
        Total observations.
    conf : float
        Confidence level that determines the width of the interval.
    bounded : {'both', 'below', 'above'}
        Which sides of the interval to close.

    Returns
    -------
    (float, float)

    References
    ----------
    .. [1]  Brown, L. D. Cai, T. T. and DasGupta, A. (2001).
            "Interval estimation for a binomial proportion".
            Statistical Science, 16(2), 101-133.
    """
    alpha = 1 - conf
    if np.any(nobs < 40):
        msg = (
            'The "agresti-coull" method is recommended when `nobs >= 40`. '
            'Consider using "wilson" or "jeffreys" methods. See [1].')
        warnings.warn(msg)
    dist = scipy.stats.norm()
    if bounded == 'both':
        z = dist.ppf(1 - alpha / 2)
        p = (count + z**2 / 2) / (nobs + z**2)
        n = nobs + z**2
        lower = np.maximum(0.0, p - z * np.sqrt(p * (1 - p) / n))
        upper = np.minimum(1.0, p + z * np.sqrt(p * (1 - p) / n))
    elif bounded == 'below':
        z = dist.ppf(1 - alpha)
        p = (count + z**2 / 2) / (nobs + z**2)
        n = nobs + z**2
        lower = np.maximum(0.0, p - z * np.sqrt(p * (1 - p) / n))
        upper = np.clip(lower, 1.0, 1.0)
    elif bounded == 'above':
        z = dist.ppf(1 - alpha)
        p = (count + z**2 / 2) / (nobs + z**2)
        n = nobs + z**2
        upper = np.minimum(1.0, p + z * np.sqrt(p * (1 - p) / n))
        lower = np.clip(upper, 0.0, 0)
    else:
        raise ValueError(util.bounded_error_msg(bounded))
    return lower, upper

def confint_1sample_jeffreys(count, nobs, conf, bounded):
    """
    Confidence interval for proportion `count / nobs`.

    Implements the Jeffreys interval, which is a bayesian method that also has
    desirable frequentist properties (see [1]_).

    Parameters
    ----------
    count : int
        Number of "true" observations.
    nobs : int
        Total observations.
    conf : float
        Confidence level that determines the width of the interval.
    bounded : {'both', 'below', 'above'}
        Which sides of the interval to close.

    Returns
    -------
    (float, float)

    References
    ----------
    .. [1]  Brown, L. D. Cai, T. T. and DasGupta, A. (2001).
            "Interval estimation for a binomial proportion".
            Statistical Science, 16(2), 101-133.
    """
    alpha = 1 - conf
    dist = scipy.stats.beta(count + 1/2, nobs - count + 1/2)
    if bounded == 'both':
        lower = dist.ppf(alpha / 2)
        upper = dist.ppf(1 - alpha / 2)
    elif bounded == 'below':
        lower = dist.ppf(alpha)
        upper = np.clip(lower, 1.0, 1.0)
    elif bounded == 'above':
        upper = dist.ppf(1 - alpha)
        lower = np.clip(upper, 0.0, 0.0)
    else:
        raise ValueError(util.bounded_error_msg(bounded))
    lower = clip_where(lower, 0.0, 0.0, np.isclose(count, 0))
    upper = clip_where(upper, 1.0, 1.0, np.isclose(count, nobs))
    return lower, upper

def confint_1sample_wilson(count, nobs, conf, bounded):
    """
    Confidence interval for proportion `count / nobs`.

    Implements the Wilson method without continuity correction (method 3 in [1]_).

    Parameters
    ----------
    count : int
        Number of "true" observations.
    nobs : int
        Total observations.
    conf : float
        Confidence level that determines the width of the interval.
    bounded : {'both', 'below', 'above'}
        Which sides of the interval to close.

    Returns
    -------
    (float, float)

    References
    ----------
    .. [1]  Newcombe, R. G. (1998).
            Two‐sided confidence intervals for the single proportion:
            comparison of seven methods.
            Statistics in medicine, 17(8), 857-872.
    """
    alpha = 1 - conf
    dist = scipy.stats.norm()
    p = count / nobs

    if bounded == 'both':
        z = dist.ppf(1 - alpha / 2)
        centre = p + z**2 / 2 / nobs
        width = z * np.sqrt(4 * nobs * p * (1 - p) + z**2) / 2 / nobs
        scale = 1 + z**2 / nobs
        lower = (centre - width) / scale
        upper = (centre + width) / scale
    elif bounded == 'below':
        z = dist.ppf(1 - alpha)
        centre = p + z**2 / 2 / nobs
        width = z * np.sqrt(4 * nobs * p * (1 - p) + z**2) / 2 / nobs
        scale = 1 + z**2 / nobs
        lower = (centre - width) / scale
        upper = np.clip(lower, 1.0, 1.0)
    elif bounded == 'above':
        z = dist.ppf(1 - alpha)
        centre = p + z**2 / 2 / nobs
        width = z * np.sqrt(4 * nobs * p * (1 - p) + z**2) / 2 / nobs
        scale = 1 + z**2 / nobs
        upper = (centre + width) / scale
        lower = np.clip(upper, 0.0, 0.0)
    else:
        raise ValueError(util.bounded_error_msg(bounded))
    lower = clip_where(lower, 0.0, 0.0, np.isclose(p, 0))
    upper = clip_where(upper, 1.0, 1.0, np.isclose(p, 1))
    return lower, upper

def confint_1sample_wilson_cc(count, nobs, conf, bounded):
    """
    Confidence interval for proportion `count / nobs`.

    Implements the Wilson method with continuity correction (method 4 in [1]_).

    Parameters
    ----------
    count : int
        Number of "true" observations.
    nobs : int
        Total observations.
    conf : float
        Confidence level that determines the width of the interval.
    bounded : {'both', 'below', 'above'}
        Which sides of the interval to close.

    Returns
    -------
    (float, float)

    References
    ----------
    .. [1]  Newcombe, R. G. (1998).
            Two‐sided confidence intervals for the single proportion:
            comparison of seven methods. Statistics in medicine, 17(8), 857-872.
    """
    alpha = 1 - conf
    dist = scipy.stats.norm()
    p = count / nobs
    if bounded == 'both':
        z = dist.ppf(1 - alpha / 2)
        a = 2 * nobs * p + z**2
        b_lower = 1 + z * np.sqrt(z**2 + 4 * nobs * p * (1 - p) - 1 / nobs + 4 * p - 2)
        b_upper = 1 + z * np.sqrt(z**2 + 4 * nobs * p * (1 - p) - 1 / nobs - 4 * p + 2)
        lower = np.maximum(0, (a - b_lower) / 2 / (nobs + z**2))
        upper = np.minimum(1, (a + b_upper) / 2 / (nobs + z**2))
    elif bounded == 'below':
        z = dist.ppf(1 - alpha)
        a = 2 * nobs * p + z**2
        b_lower = 1 + z * np.sqrt(z**2 + 4 * nobs * p * (1 - p) - 1 / nobs + 4 * p - 2)
        lower = np.maximum(0, (a - b_lower) / 2 / (nobs + z**2))
        upper = np.clip(lower, 1.0, 1.0)
    elif bounded == 'above':
        z = dist.ppf(1 - alpha)
        a = 2 * nobs * p + z**2
        b_upper = 1 + z * np.sqrt(z**2 + 4 * nobs * p * (1 - p) - 1 / nobs - 4 * p + 2)
        upper = np.minimum(1, (a + b_upper) / 2 / (nobs + z**2))
        lower = np.clip(upper, 0.0, 0.0)
    else:
        raise ValueError(util.bounded_error_msg(bounded))
    lower = clip_where(lower, 0.0, 0.0, np.isclose(p, 0))
    upper = clip_where(upper, 1.0, 1.0, np.isclose(p, 1))
    return lower, upper

def confint_2sample_agresti_caffo(count1, nobs1, count2, nobs2, conf, bounded):
    """
    Confidence interval for difference between proportions `count1 / nobs1 - count2 / nobs2`.

    Implements the Agresti-Caffo method. Note that this method performs well
    over most parameters, but is not optimal over all parameters. For details
    see ([1]_).

    Parameters
    ----------
    count1 : int
        Number of "true" observations in first sample.
    nobs1 : int
        Total observations in first sample.
    count2 : int
        Number of "true" observations in second sample.
    nobs2 : int
        Total observations in second sample.
    conf : float
        Confidence level that determines the width of the interval.
    bounded : {'both', 'below', 'above'}
        Which sides of the interval to close.

    Returns
    -------
    :class:`mqr.inference.confint.ConfidenceInterval`

    References
    ----------
    .. [1]  Agresti, A., & Caffo, B. (2000).
            Simple and effective confidence intervals for proportions and differences
            of proportions result from adding two successes and two failures.
            The American Statistician, 54(4), 280-288.
    """
    alpha = 1 - conf
    p1 = (count1 + 1) / (nobs1 + 2)
    p2 = (count2 + 1) / (nobs2 + 2)
    mu = p1 - p2
    sigma = np.sqrt(p1 * (1 - p1) / (nobs1 + 2) + p2 * (1 - p2) / (nobs2 + 2))
    dist = scipy.stats.norm(mu, sigma)
    if bounded == 'both':
        lower = dist.ppf(alpha / 2)
        upper = dist.ppf(1 - alpha / 2)
    elif bounded == 'below':
        lower = dist.ppf(alpha)
        upper = np.clip(lower, np.inf, np.inf)
    elif bounded == 'above':
        upper = dist.ppf(1 - alpha)
        lower = np.clip(upper, -np.inf, -np.inf)
    else:
        raise ValueError(util.bounded_error_msg(bounded))
    return lower, upper

def confint_2sample_newcomb(count1, nobs1, count2, nobs2, conf, bounded):
    """
    Confidence interval for difference between proportions `count1 / nobs1 - count2 / nobs2`.

    Implements method 10 in [1]_.

    Parameters
    ----------
    count1 : int
        Number of "true" observations in first sample.
    nobs1 : int
        Total observations in first sample.
    count2 : int
        Number of "true" observations in second sample.
    nobs2 : int
        Total observations in second sample.
    conf : float
        Confidence level that determines the width of the interval.
    bounded : {'both', 'below', 'above'}
        Which sides of the interval to close.

    Returns
    -------
    :class:`mqr.inference.confint.ConfidenceInterval`

    References
    ----------
    .. [1]  Newcombe, R. G. (1998).
            Interval estimation for the difference between independent proportions:
            comparison of eleven methods.
            Statistics in medicine, 17(8), 873-890.
    """
    alpha = 1 - conf
    theta_hat = count1 / nobs1 - count2 / nobs2
    l1, u1 = confint_1sample_wilson(count1, nobs1, conf, bounded)
    l2, u2 = confint_1sample_wilson(count2, nobs2, conf, bounded)

    if bounded == 'both':
        z = scipy.stats.norm().ppf(1 - alpha / 2)
        delta = z * np.sqrt(l1 * (1 - l1) / nobs1 + u2 * (1 - u2) / nobs2)
        epsilon = z * np.sqrt(u1 * (1 - u1) / nobs1 + l2 * (1 - l2) / nobs2)
        lower = theta_hat - delta
        upper = theta_hat + epsilon
    elif bounded == 'below':
        z = scipy.stats.norm().ppf(1 - alpha)
        delta = z * np.sqrt(l1 * (1 - l1) / nobs1 + u2 * (1 - u2) / nobs2)
        lower = theta_hat - delta
        upper = np.clip(lower, np.inf, np.inf)
    elif bounded == 'above':
        z = scipy.stats.norm().ppf(1 - alpha)
        epsilon = z * np.sqrt(u1 * (1 - u1) / nobs1 + l2 * (1 - l2) / nobs2)
        upper = theta_hat + epsilon
        lower = np.clip(upper, -np.inf, -np.inf)
    else:
        raise ValueError(util.bounded_error_msg(bounded))
    return lower, upper

def confint_2sample_newcomb_cc(count1, nobs1, count2, nobs2, conf, bounded):
    """
    Confidence interval for difference between proportions `count1 / nobs1 - count2 / nobs2`.

    Implements method 11 in [1]_. Further discussion and comparisons in [2]_.

    Parameters
    ----------
    count1 : int
        Number of "true" observations in first sample.
    nobs1 : int
        Total observations in first sample.
    count2 : int
        Number of "true" observations in second sample.
    nobs2 : int
        Total observations in second sample.
    conf : float
        Confidence level that determines the width of the interval.
    bounded : {'both', 'below', 'above'}
        Which sides of the interval to close.

    Returns
    -------
    :class:`mqr.inference.confint.ConfidenceInterval`

    References
    ----------
    .. [1]  Newcombe, R. G. (1998).
            Interval estimation for the difference between independent proportions:
            comparison of eleven methods.
            Statistics in medicine, 17(8), 873-890.
    .. [2]  Agresti, A., & Caffo, B. (2000).
            Simple and effective confidence intervals for proportions and differences
            of proportions result from adding two successes and two failures.
            The American Statistician, 54(4), 280-288.
    """
    p1 = count1 / nobs1
    p2 = count2 / nobs2
    if bounded == 'both':
        lower1, upper1 = confint_1sample_wilson_cc(count1, nobs1, conf, bounded)
        lower2, upper2 = confint_1sample_wilson_cc(count2, nobs2, conf, bounded)
        lower = (p1 - p2) - np.sqrt((p1 - lower1)**2 + (upper2 - p2)**2)
        upper = (p1 - p2) + np.sqrt((upper1 - p1)**2 + (p2 - lower2)**2)
    elif bounded == 'below':
        lower1, _ = confint_1sample_wilson_cc(count1, nobs1, conf, 'below')
        _, upper2 = confint_1sample_wilson_cc(count2, nobs2, conf, 'above')
        lower = (p1 - p2) - np.sqrt((p1 - lower1)**2 + (upper2 - p2)**2)
        upper = np.clip(lower, np.inf, np.inf)
    elif bounded == 'above':
        _, upper1 = confint_1sample_wilson_cc(count1, nobs1, conf, 'above')
        lower2, _ = confint_1sample_wilson_cc(count2, nobs2, conf, 'below')
        upper = (p1 - p2) + np.sqrt((upper1 - p1)**2 + (p2 - lower2)**2)
        lower = np.clip(upper, -np.inf, -np.inf)
    else:
        raise ValueError(util.bounded_error_msg(bounded))
    return lower, upper



"""
Fishbone/Ishikawa diagrams for modelling root-causes.
"""

import matplotlib.pyplot as plt
import matplotlib.transforms as transforms
import numpy as np

from mqr.plot.lib.util import set_kws
from mqr.plot.defaults import Defaults

def ishikawa(problem:str, causes:dict[str,list[str]], ax, ishikawa_kws=None):
    """
    Draws an Ishikawa/fishbone diagram.

    Draws the "spine", and other levels of "bones". Express the label for the spine
    as the `problem` paramerter. Express the two levels of causes as a dictionary
    of lists: the key to each dictionary names the first branch from the spine,
    the list of values name secondary branches.

    Parameters
    ----------
    problem : str
        Label for the main spine.
    causes : dict[str, list[str]]
        Mapping from string values which are labels for the first branches to
        list-of-string keys which are labels for the next level causes.
    ax : matplotlib.axes.Axes
        Axes for the plot.
    ishikawa_kws : dict, optional
        Keys overriding :attr:`mqr.plot.defaults.Defaults.ishikawa`.

    Examples
    --------
    This example shows a diagram with the main spine labelled as "Problem". Two
    branches leave the spine, then other branches leave those.

    .. plot::
        :include-source:

        import mqr

        fig, ax = plt.subplots(figsize=(5, 3))

        problem = 'Problem'
        causes = {
            'Material':    ['Cause1a', 'Cause1b', 'Cause1c'],
            'People':      ['Cause2a', 'Cause2b'],
            'Machine':     ['Cause3a', 'Cause3b', 'Cause3c'],
            'Measurement': ['Cause4a', 'Cause4b', 'Cause4c', 'Cause4d'],
            'Method':      ['Cause5a', 'Cause5b', 'Cause5c', 'Cause5c'],
            'Environment': ['Cause6a', 'Cause6b', 'Cause6c'],
        }
        mqr.plot.ishikawa(problem, causes, ax=ax)

        plt.show()

    """
    if ishikawa_kws is not None:
        conf = Defaults.ishikawa | ishikawa_kws
    else:
        conf = Defaults.ishikawa

    pad_t, pad_r, pad_b, pad_l = conf['padding']

    bones = list(causes.keys())
    max_causes = max([len(v) for v in causes.values()])
    N = int(np.ceil(len(bones) / 2))

    bone_rise = conf['bone_rise'] = (max_causes + 1) * conf['cause_space']
    head_space = conf['head_space']
    bone_run = conf['bone_run'] = conf['bone_rise'] / np.tan(conf['bone_angle'])
    bone_space = conf['bone_space']
    body_length = (N - 1) * bone_space

    _draw_spine(body_length, head_space, conf=conf, ax=ax)
    ax.text(head_space + 0.2, 0, problem,
            **conf['defect_font_dict'], va='center', ha='left')

    for i in range(N):
        _draw_bone(
            bones[2*i],
            x_start=-i*bone_space,
            bone_rise=bone_rise,
            bone_run=bone_run,
            bone_drn=1,
            conf=conf,
            ax=ax)
        top_causes = causes[bones[2*i]]
        cause_start = (max_causes - len(top_causes)) / 2
        for j, cause in enumerate(top_causes):
            _draw_primary_cause(cause, i, 1, cause_start+j, conf=conf, ax=ax)

        bone_text = '' if (len(bones) <= 2*i+1) else bones[2*i+1]
        _draw_bone(
            bone_text,
            x_start=-i*bone_space,
            bone_rise=bone_rise,
            bone_run=bone_run,
            bone_drn=-1,
            conf=conf,
            ax=ax)

        try:
            bottom_causes = causes[bones[2*i+1]]
            cause_start = (max_causes - len(bottom_causes)) / 2
            for j, cause in enumerate(bottom_causes):
                _draw_primary_cause(cause, i, -1, cause_start+j, conf=conf, ax=ax)
        except:
            pass

    ax.set_aspect('equal', 'box')
    ax.axis('off')

def _draw_spine(body_length, head_space, *, conf, ax):
    x_left, x_right = -body_length, head_space
    y = 0.0
    ax.plot([x_left, x_right], [y, y], **conf['line_kwargs'])

def _draw_bone(text, x_start, bone_rise, bone_run, bone_drn, *, conf, ax):
    x_end = x_start - bone_run
    y_start, y_end = 0.0, bone_drn * bone_rise
    ax.plot([x_start, x_end], [y_start, y_end], **conf['line_kwargs'])
    text_offset = 0.2 * bone_drn
    va = 'bottom' if bone_drn == 1 else 'top'
    ax.text(x_end, y_end+text_offset, text, conf['cause_font_dict'], va=va, ha='center')

def _draw_primary_cause(text, bone_index, bone_drn, height, *, conf, ax):
    bone_start = -bone_index * conf['bone_space']
    y = (height + 1) * conf['cause_space']
    x_right = bone_start - y * np.tan(conf['bone_angle'])
    x_left = bone_start - y * np.tan(conf['bone_angle']) - conf['cause_length']

    ax.annotate(text,
                xy=(x_right, bone_drn * y),
                ha='right',
                va='center',
                xytext=(x_left, bone_drn * y),
                **conf['primary_font_dict'],
                xycoords='data',
                textcoords='data',
                arrowprops=dict(arrowstyle="->",
                                facecolor='black'))



"""
============================================
Analysis of variance (:mod:`mqr.plot.anova`)
============================================

.. currentmodule:: mqr.plot.anova

.. rubric:: Functions
.. autosummary::
    :toctree: generated/

    main_effects
    interactions
    model_means
"""

from collections.abc import Iterable
import numpy as np

import mqr
from mqr.plot.defaults import Defaults
from mqr.plot.lib.util import set_kws

def groups(groups_df, ax, ci_kws=None):
    """
    Draw a bar graph with error bars for the groups in an ANOVA.

    Parameters
    ----------
    ci : :class:`mqr.confint.ConfidenceInterval`
        The confidence interval to draw.
    ax : matplotlib.axes.Axes
        Axes for the plot.
    """
    ci_kws = set_kws(
        ci_kws,
        fmt=Defaults.marker,
        capsize=4.0,
    )

    y_err = (groups_df.iloc[:, -1] - groups_df.iloc[:, -2]) / 2
    ax.errorbar(
        x=groups_df.index,
        y=groups_df['mean'],
        yerr=y_err,
        **ci_kws)
    ax.set_xticks(groups_df.index)
    ax.set_xmargin(0.2)
    ax.grid(alpha=0.5)

def main_effects(data, response, factors, *, axs, line_kws=None, mean_kws=None):
    """
    Plot the main effects from experimental data.

    Parameters
    ----------
    data : pandas.DataFrame
    response : str
    factors : list[str]
    axs : array_like[matplotlib.axes.Axes]
        Axes for the plot. Must have the same number of elements as `factors`.
        Will be flattened before use.
    line_kws : dict, optional
        Keyword args for the effect lines.
        Passed to :func:`matplotlib.pyplot.plot`.
    mean_kws : dict, optional
        Keyword args for the overall average line.
        Passed to :func:`matplotlib.pyplot.axhline`.

    Examples
    --------
    This example loads sample data and shows a main effects plot for the two factors.

    .. plot::

        data = pd.read_csv(mqr.sample_data('anova-glue.csv'), index_col='Run')

        fig, axs = plt.subplots(1, 2, figsize=(5, 2), layout='constrained')
        mqr.plot.tools.sharey(fig, axs)
        mqr.plot.anova.main_effects(
            data,
            response='adhesion_force',
            factors=['primer', 'glue'],
            axs=axs)

    """
    axs = axs.flatten()
    if len(factors) != len(axs):
        raise ValueError('Number of axes must equal the length of `factors`.')

    line_kws = set_kws(
        line_kws,
        marker=Defaults.marker,
    )
    mean_kws = set_kws(
        mean_kws,
        linestyle='--',
        color='k',
        alpha=0.6,
    )

    mean = data[response].mean()
    axs[0].set_ylabel(response)
    for ax, factor in zip(axs, factors):
        grp = data.groupby(factor)[response].mean()
        ax.plot(grp, **line_kws)
        ax.axhline(mean, **mean_kws)
        ax.set_xticks(data[factor].unique())
        ax.set_xmargin(0.2)
        ax.set_xlabel(factor)

def interactions(data, response, group, factors, *, axs, line_kws=None, mean_kws=None):
    """
    Interaction plot for observation `obs` between the two categories in `between`.
    
    Parameters
    ----------
    data : pd.DataFrame
        Table of categorical data.
    response : str
        Column name that contains response (real numbers).
    group : str
        Categorical factor whose values are shown as separate lines on each plot.
    factors : list[str]
        Categorical factors whose interactions with `group` are plot on each axis.
    axs : matplotlib.axes.Axes, list[matplotlib.axes.Axes]
        Axes for the plot. Must match the length of `factors` (if the length of
        `factors` is 1, then `axs` can be an axes object).
    line_kws : dict, optional
        Keyword arguments for the lines. Passed to :func:`matplotlib.pyplot.plot`.
    mean_kws : dict, optional
        Keyword arguments for the mean line.
        Passed to :func:`matplotlib.pyplot.axhline`.

    Examples
    --------
    This example loads sample data and shows an interaction plot for the two factors.
    Since one factor is used for grouping, there is one factor left for plots.

    .. plot::

        data = pd.read_csv(mqr.sample_data('anova-glue.csv'), index_col='Run')

        fig, axs = plt.subplots(figsize=(3, 2), layout='constrained')
        mqr.plot.anova.interactions(
            data,
            response='adhesion_force',
            group='primer',
            factors=['glue'],
            axs=axs)

    """

    if not isinstance(axs, Iterable):
        axs = np.array([axs])

    axs = axs.flatten()
    if len(factors) != len(axs):
        raise ValueError('Number of axes must equal the length of `factors`.')

    line_kws = set_kws(
        line_kws,
        marker=Defaults.marker,
    )
    mean_kws = set_kws(
        mean_kws,
        linestyle='--',
        color='k',
        alpha=0.6,
    )

    mean = data[response].mean()
    axs[0].set_ylabel(response)
    levels = data[group].unique()
    for ax, factor in zip(axs, factors):
        for level in levels:
            slice = data[group] == level
            values = data[slice].groupby(factor)[response].mean()
            ax.plot(values, **line_kws)
        ax.axhline(mean, **mean_kws)
        ax.set_xticks(data[factor].unique())
        ax.set_xmargin(0.2)
        ax.set_xlabel(factor)
        ax.legend(levels, title=group)

def model_means(result, response, factors, axs, ci_kws=None):
    """
    Plot the means of each combination of factor levels in an ANOVA result.

    Parameters
    ----------
    result : statsmodels.regression.linear_model.RegressionResults
        Result of calling `fit` on a statsmodel linear regression model.
    response : str
        Name of response variable.
    factors : list[str]
        List of names of categorical factors.
    axs : array_like[matplotlib.axes.Axes]
        Axes for the plot. Must have the same number of elements as `factors`.
        Will be flattened before use.
    ci_kws: dict, optional
        Keyword args for confidence intervals.
        Passed to :func:`matplotlib..errorbar <matplotlib.pyplot.errorbar>`.

    Examples
    --------
    This example performs an ANOVA on sample data, then shows the means of the
    factors.

    .. plot::
        :include-source:

        from statsmodels.formula.api import ols

        data = pd.read_csv(mqr.sample_data('anova-glue.csv'), index_col='Run')
        model = ols('adhesion_force ~ C(primer) + C(glue)', data=data)
        result = model.fit()

        fig, axs = plt.subplots(1, 2, figsize=(4, 2), layout='constrained')
        mqr.plot.tools.sharey(fig, axs)
        mqr.plot.anova.model_means(
            result,
            response='adhesion_force',
            factors=['primer', 'glue'],
            axs=axs)

    """
    axs = axs.flatten()
    if len(axs) != len(factors):
        raise ValueError('Length of `axs` must equal length of `factors`.')

    for ax, factor in zip(axs, factors):
        df_grp = mqr.anova.groups(result, value=response, factor=factor, formatted=False)
        mqr.plot.anova.groups(df_grp, ax, ci_kws)
        ax.set_xlabel(factor)



from dataclasses import dataclass, field
import numpy as np
import scipy

import mqr.inference.lib.util as util

def _hyptest_table_styles():
        return [
            {
                'selector': '.row_heading',
                'props': [
                    ('text-align', 'left'),
                ]
            },
        ]

@dataclass
class HypothesisTest:
    """
    Result of an hypothesis test.

    Attributes
    ----------
    description : str
        Description of the statistic on which the test is performed.
    alternative : {'two-sided', 'less', 'greater'}
        Sense of the alternative hypothesis.
    method : str
        Name of the test method, eg. 'Kruskall-Wallace'.
    sample_stat : str
        Name of the statistic from the sample on which the test is performed.
        For example, when performing an hypothesis test on the mean of a sample,
        `sample_stat` is the mean. This is independent of the method.
    sample_stat_target : str
        Hypothesised value of the sample statistic. For example, when performing
        an hypothesis test on the mean of a sample, `sample_stat_target` is
        the hypothesised mean that appears in the null-hypothesis. Expressed as
        a string because some hypothesis tests target non-numeric properties,
        like distributions.
    sample_stat_value : float
        Actual value of the sample statistic calculated from the sample. Use
        `None` when the `sample_stat_target` is non-numeric.
    stat : float
        Test statistic. For example, in an hypothesis test on the mean of a
        sample, `stat` might be the score on a student's-t distribution.
    pvalue : float
        p-value associated with the test statistic.
    null : None
        Automatically generated. A string representation of the null-hypothesis.
    alt : None
        Automatically generated. A string representation of the alternative-hypothesis.

    Notes
    -----
    This class is iterable, and iterates over the test statistic and p-value (in
    that order).

    Examples
    --------
    All hypothesis tests return a type constructed like this.

    >>> import mqr
    >>> hyptest = mqr.inference.hyptest.HypothesisTest(
    >>>     description='Example mean test',
    >>>     alternative='less',
    >>>     method='Example students-t',
    >>>     sample_stat='mean(x)',
    >>>     sample_stat_target='1.6',
    >>>     sample_stat_value=1.4,
    >>>     stat=-2.828427124746192,
    >>>     pvalue=0.02371032779215975)
    >>> hyptest
    HypothesisTest(
        description='Example mean test',
        alternative='less',
        method='Example students-t',
        sample_stat='mean(x)',
        sample_stat_target='1.6',
        sample_stat_value=1.4,
        stat=-2.828427124746192,
        pvalue=0.02371032779215975,
        null='mean(x) == 1.6',
        alt='mean(x) < 1.6')

    Iterate over the test statistic and p-value:

    >>> stat, pvalue = hyptest
    >>> stat, pvalue
    (-2.828427124746192, 0.02371032779215975)

    When displayed in a jupyter notebook, the example here will be shown as an
    HTML table:

    +------------------------+---------------------+
    | | **Hypothesis Test**                        |
    | | Example mean test                          |
    +========================+=====================+
    | **method**             | Example students-t  |
    +------------------------+---------------------+
    | **H**:sub:`0`          | mean(x) == 1.6      |
    +------------------------+---------------------+
    | **H**:sub:`1`          | mean(x) < 1.6       |
    +------------------------+---------------------+
    | **statistic**          | -2.828427124746192  |
    +------------------------+---------------------+
    | **p-value**            | 0.02371032779215975 |
    +------------------------+---------------------+
    """
    description: str
    alternative: str
    method: str
    sample_stat: str
    sample_stat_target: str
    sample_stat_value: np.float64
    stat: np.float64
    pvalue: np.float64

    null: str = None
    alt: str = None

    def __post_init__(self):
        self.null = self._null_hypothesis()
        self.alt = self._alt_hypothesis()

    def __iter__(self):
        """
        Iterator over the test statistic and p-value.
        """
        return iter((self.stat, self.pvalue))
        
    def _null_hypothesis(self):
        import numbers
        if isinstance(self.sample_stat_target, numbers.Number):
            fmt = 'g'
        else:
            fmt = 's'
        return f'{self.sample_stat} == {self.sample_stat_target:{fmt}}'

    def _alt_hypothesis(self):
        import numbers

        alt_sym = None
        if self.alternative == 'two-sided':
            alt_sym = '!='
        elif self.alternative == 'less':
            alt_sym = '<'
        elif self.alternative == 'greater':
            alt_sym = '>'
        else:
            raise RuntimeError(util.alternative_error_msg(self.alternative))

        if isinstance(self.sample_stat_target, numbers.Number):
            fmt = 'g'
        else:
            fmt = 's'

        return f'{self.sample_stat} {alt_sym} {self.sample_stat_target:{fmt}}'

    def _html(self):
        return f'''
        <table>
        <thead>
            <tr>
                <th scope="col" colspan=2 style="text-align: left; padding-bottom: 0px;">Hypothesis Test</th>
            </tr>
            <tr style='padding-top: 0px;'>
                <td colspan=2 style='text-align: left; padding-top: 0px'>{self.description}</td>
            </tr>
        </thead>
        <tbody>
            <tr>
                <th scope="row">method</th>
                <td>{self.method}</th>
            </tr>
            <tr>
                <th scope="row">H<sub>0</sub></th>
                <td>{self.null}</td>
            </tr>
            <tr>
                <th scope="row">H<sub>1</sub></th>
                <td>{self.alt}</td>
            </tr>
            <thead><tr/></thead>
            <tr>
                <th scope="row">statistic</th>
                <td>{self.stat:g}</td>
            </tr>
            <tr>
                <th scope="row">p-value</th>
                <td>{self.pvalue:g}</td>
            </tr>
        </tbody>
        </table>
        '''

    def _repr_html_(self):
        return self._html()



from mqr.inference.confint import ConfidenceInterval
from mqr.inference.hyptest import HypothesisTest

import mqr.inference.lib.util as util
import mqr.interop.inference as interop

import warnings

def test_1sample(x, test='ad-norm'):
    """
    Hypothesis test on sampling distribution.

    Null-hypothesis
        | 'ad-norm', 'ks-norm'
        |   `x` was sampled from the a normal distribution

    Parameters
    ----------
    x : array_like
        Test the distribution of these values.
    test : {'ad-norm', 'ks-norm'}, optional
        | 'ad-norm'
        |   Anderson-Darling normality test.
            Calls :func:`sm..normal_ad <statsmodels.stats.diagnostic.normal_ad>`.
        | 'ks-norm'
        |   Kolmogoroc-Smirnov test against the normal distribution.
            Calls :func:`sm..kstest_normal <statsmodels.stats.diagnostic.kstest_normal>`.
        | 'lf-norm'
        |   Lilliefors test against the normal distribution.
            Lilliefors test is a Kolmogorov-Smirnov test with estimated parameters.
            Calls :func:`sm..lilliefors <statsmodels.stats.diagnostic.lilliefors>`.
        | 'sw-norm'
        |   Shapiro-Wilk normality test.
            Calls :func:`scipy..shaprio <scipy.stats.shapiro>`.
        | 'dp-norm'
        |   Dagastino-Pearson normality test.
            Calls :func:`scipy..normaltest <scipy.stats.normaltest>`.
        | 'jb-norm'
        |   Jarque-Bera normality test. Can lead to high Type-I error under some conditions,
            see the discussion on `Wikipedia <https://en.wikipedia.org/wiki/Jarque–Bera_test>`_.
            Calls :func:`scipy..jarque_bera <scipy.stats.jarque_bera>`.
        | 'cvm-norm'
        |   Cramer-von Mises test for goodness of fit, applied to the standard Normal CDF.
            Calls :func:`scipy..cramervonmises <scipy.stats.cramervonmises>`.

    Returns
    -------
    :class:`mqr.inference.hyptest.HypothesisTest`
    """
    if test == 'ad-norm':
        from statsmodels.stats.diagnostic import normal_ad
        description = 'non-normality'
        alternative = 'two-sided'
        method = 'anderson-darling'
        target = 'normal'
        statistic, pvalue = normal_ad(x)
    elif test == 'ks-norm':
        from statsmodels.stats.diagnostic import kstest_normal
        description = 'non-normality'
        alternative = 'two-sided'
        method = 'kolmogorov-smirnov'
        target = 'normal'
        statistic, pvalue = kstest_normal(x, dist='norm')
    elif test == 'lf-norm':
        from statsmodels.stats.diagnostic import lilliefors
        description = 'non-normality'
        alternative = 'two-sided'
        method = 'lilliefors'
        target = 'normal'
        statistic, pvalue = lilliefors(x, dist='norm')
    elif test == 'sw-norm':
        from scipy.stats import shapiro
        description = 'non-normality'
        alternative = 'two-sided'
        method = 'shapiro-wilk'
        target = 'normal'
        statistic, pvalue = shapiro(x)
    elif test == 'dp-norm':
        from scipy.stats import normaltest
        description = 'non-normality'
        alternative = 'two-sided'
        method = 'dagostino-pearson'
        target = 'normal'
        statistic, pvalue = normaltest(x)
    elif test == 'jb-norm':
        if len(x) < 100:
            msg = (
                'This method is recommended for large samples. '
                'Consider using another method. '
                'See the discussion in https://en.wikipedia.org/wiki/Jarque–Bera_test.')
            warnings.warn(msg)
        from scipy.stats import jarque_bera
        description = 'non-normality'
        alternative = 'two-sided'
        method = 'jarque-bera'
        target = 'normal'
        statistic, pvalue = jarque_bera(x)
        if pvalue < 0.05:
            msg = (
                'This method might be inaccurate when the p-value is small. '
                'Consider using another method. '
                'See the discussion in https://en.wikipedia.org/wiki/Jarque–Bera_test.')
            warnings.warn(msg)
    elif test == 'cvm-norm':
        from scipy.stats import cramervonmises, norm
        description = 'non-normality'
        alternative = 'two-sided'
        method = 'cramer-vonmises'
        target = 'normal'
        res = cramervonmises(x, norm().cdf)
        statistic, pvalue = res.statistic, res.pvalue
    else:
        valid_methods = [
            'ad-norm',
            'ks-norm',
            'lf-norm',
            'sw-norm',
            'dp-norm',
            'jb-norm',
            'cvm-norm',
        ]
        raise ValueError(util.method_error_msg(test, valid_methods))

    return HypothesisTest(
        description=description,
        alternative=alternative,
        method=method,
        sample_stat=f'distribution',
        sample_stat_target=target,
        sample_stat_value=None,
        stat=statistic,
        pvalue=pvalue,
    )



'''
Check result types and utils.
'''

import numbers
import numpy as np
import pytest

import mqr

def test_alternative():
    assert mqr.interop.inference.alternative('two-sided', 'statsmodels') == 'two-sided'
    assert mqr.interop.inference.alternative('less', 'statsmodels') == 'smaller'
    assert mqr.interop.inference.alternative('greater', 'statsmodels') == 'larger'

def test_bounded():
    assert mqr.interop.inference.bounded('both', 'scipy') == 'two-sided'
    assert mqr.interop.inference.bounded('below', 'scipy') == 'greater'
    assert mqr.interop.inference.bounded('above', 'scipy') == 'less'

    assert mqr.interop.inference.bounded('both', 'statsmodels') == 'two-sided'
    assert mqr.interop.inference.bounded('below', 'statsmodels') == 'larger'
    assert mqr.interop.inference.bounded('above', 'statsmodels') == 'smaller'

def test_confint_result():
    conf = mqr.inference.confint.ConfidenceInterval(
        name='',
        method='',
        value=np.nan,
        lower=-1.234,
        upper=1.234,
        conf=np.nan,
        bounded='both')

    (lower, upper) = conf
    assert lower == -1.234
    assert upper == 1.234

def test_hyptest_result():
    description = 'description-str'
    method = 'method-str'
    sample_stat = 'sample_stat-str'
    sample_stat_target = 'sample_stat_target-str'
    sample_stat_value = 1.234
    stat = 12.34
    pvalue = 0.1234

    alternative = 'two-sided'
    res = mqr.inference.hyptest.HypothesisTest(
        description=description,
        alternative=alternative,
        method=method,
        sample_stat=sample_stat,
        sample_stat_target=sample_stat_target,
        sample_stat_value=sample_stat_value,
        stat=stat,
        pvalue=pvalue)
    assert res.null == 'sample_stat-str == sample_stat_target-str'
    assert res.alt == 'sample_stat-str != sample_stat_target-str'

    alternative = 'less'
    res = mqr.inference.hyptest.HypothesisTest(
        description=description,
        alternative=alternative,
        method=method,
        sample_stat=sample_stat,
        sample_stat_target=sample_stat_target,
        sample_stat_value=sample_stat_value,
        stat=stat,
        pvalue=pvalue)
    assert res.null == 'sample_stat-str == sample_stat_target-str'
    assert res.alt == 'sample_stat-str < sample_stat_target-str'

    alternative = 'greater'
    res = mqr.inference.hyptest.HypothesisTest(
        description=description,
        alternative=alternative,
        method=method,
        sample_stat=sample_stat,
        sample_stat_target=sample_stat_target,
        sample_stat_value=sample_stat_value,
        stat=stat,
        pvalue=pvalue)
    assert res.null == 'sample_stat-str == sample_stat_target-str'
    assert res.alt == 'sample_stat-str > sample_stat_target-str'



'''
Check call-throughs.
'''

import numbers
import numpy as np
import pytest
import scipy

import mqr

def test_confint_1sample():
    x = np.array([0.36522057, 0.37377119, 0.86150726, 0.96718967, 0.47966424,
        0.74979588, 0.35591634, 0.75272332, 0.17909715, 0.66216129])
    q = 0.2
    conf = 0.90
    bounded = 'below'

    res = mqr.inference.nonparametric.quantile.confint_1sample(x, q, conf, bounded)
    assert res.name == 'quantile (20th percentile)'
    assert res.method == 'binom'
    assert isinstance(res.value, numbers.Number)
    assert isinstance(res.lower, numbers.Number)
    assert res.upper == np.inf
    assert res.conf == conf
    assert res.bounded == bounded

def test_test_1sample():
    x = np.array([0.36522057, 0.37377119, 0.86150726, 0.96718967, 0.47966424,
        0.74979588, 0.35591634, 0.75272332, 0.17909715, 0.66216129])
    H0_quant = np.quantile(x, 0.2) + 0.1
    q = 0.2
    alternative = 'greater'

    res = mqr.inference.nonparametric.quantile.test_1sample(x, H0_quant, q, alternative)
    assert res.description == 'quantile'
    assert res.alternative == alternative
    assert res.method == 'binom'
    assert res.sample_stat == 'quantile(x, 0.2)'
    assert res.sample_stat_target == H0_quant
    assert res.sample_stat_value == np.quantile(x, q)
    assert isinstance(res.stat, numbers.Number)
    assert isinstance(res.pvalue, numbers.Number)

    H0_quant = None
    res = mqr.inference.nonparametric.quantile.test_1sample(x, H0_quant, q, alternative)
    assert res.description == 'quantile'
    assert res.alternative == alternative
    assert res.method == 'binom'
    assert res.sample_stat == 'quantile(x, 0.2)'
    assert res.sample_stat_target == np.quantile(x, q)
    assert res.sample_stat_value == np.quantile(x, q)
    assert isinstance(res.stat, numbers.Number)
    assert isinstance(res.pvalue, numbers.Number)



"""
=======================================
Analysis of Variance (:mod:`mqr.anova`)
=======================================

.. :currentmodule:: mqr.anova

User guide
    :doc:`/user_guide/regression-anova`

Detailed examples
    https://github.com/nklsxn/mqrpy-guide

Tools for interpreting ANOVA results.

.. rubric:: Functions

.. autosummary::
    :toctree: generated/

    adequacy
    summary
    coeffs
    groups
    interactions
"""

from dataclasses import dataclass
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import scipy
import scipy.stats as st
import seaborn as sns
import statsmodels
import statsmodels.api as sm

from statsmodels.stats.outliers_influence import variance_inflation_factor

import warnings

################################################################################
## Tools
################################################################################

def summary(result, typ=2, formatted=True):
    '''
    The ANOVA table for a regression.

    Parameters
    ----------
    result : statsmodels.regression.linear_model.RegressionResults
        Result of calling `fit` on a statsmodels linear regression model.
    typ : {1, 2, 3, 'I', 'II', 'III'}, optional
        The ANOVA analysis type. Passed to :func:`anova_lm(..., typ=typ) 
        <statsmodels.stats.anova.anova_lm>`.
    formatted : bool, optional
        When True, returns a formatted DataFrame. Set this to False to access
        the calculated values.

    Returns
    -------
    pandas.DataFrame or pandas.io.formats.style.Styler
        The ANOVA table for the regression. A DataFrame when `formatted` is False
        and a `Styler` when True.
    '''
    table = sm.stats.anova_lm(result, typ=typ)
    table.loc['Total'] = table.sum(axis=0, skipna=False)
    table['mean_sq'] = table['sum_sq'] / table['df']
    table = table[['df', 'sum_sq', 'mean_sq', 'F', 'PR(>F)']]

    if formatted:
        format_map = {
            'df': '{:.0f}',
            'sum_sq': '{:,.4g}',
            'mean_sq': '{:.4g}',
            'F': '{:.2f}',
            'PR(>F)': '{:.3f}',
        }
        return table.style.format(format_map)
    else:
        return table

def coeffs(result, conf=0.95, formatted=True):
    """
    The coefficients from regression and their confidence intervals.

    Parameters
    ----------
    result : statsmodels.regression.linear_model.RegressionResults
        Result of calling `fit` on a statsmodels linear regression model.
    conf: float, optional
        Confidence level used to form an interval (decimal).
    formatted : bool, optional
        When True, returns a formatted DataFrame. Set this to False to access
        the calculated values.

    Returns
    -------
    pandas.DataFrame or pandas.io.formats.style.Styler
        Coefficients indexed by name from the model. A DataFrame when `formatted`
        is True, a Styler otherwise.
    """
    alpha = 1 - conf
    values = pd.concat(
        [result.params, result.conf_int(alpha=alpha)],
        axis=1)
    values.columns = ['Coeff', 'lower', 'upper']
    values['t'] = result.tvalues
    values['PR(>|t|)'] = result.pvalues
    exog = result.model.exog
    if exog.shape[1] > 1:
        values['VIF'] = np.nan
        for i in np.arange(exog.shape[1]):
            values.iloc[i, -1] = variance_inflation_factor(exog, i)

    if formatted:
        format_map = {
            'Coeff': '{:.4g}',
            'lower': '{:,.4g}',
            'upper': '{:.4g}',
            't': '{:.2f}',
            'PR(>|t|)': '{:.3f}',
            'VIF': '{:.3g}',
        }
        return values.style.format(format_map)
    else:
        return values

def groups(result, *, value: str, factor: str, conf=0.95, formatted=True):
    """
    The `value` column from a dataframe averaged over `factor`, and annotated
    with confidence intervals per level.

    See [1]_ (pp. 538-541) for construction of t-statistic.

    Parameters
    ----------
    df : pandas.DataFrame
        A dataframe of measurements and categories.
    value : str
        Name of the column to average into groups.
    factor : str
        Name of the column (categorical) to group by.
    conf : float, optional
        The confidence level used to form an interval (decimal).
    formatted : bool, optional
        When True, returns a formatted DataFrame. Set this to False to access
        the calculated values.

    Returns
    -------
    pandas.DataFrame or pandas.io.formats.style.Styler
        Average values per group with confidence intervals. A DataFrame when
        `formatted` is True, a Styler otherwise.

    References
    ----------
    .. [1]  Saville, David J., and Graham R. Wood.
            Statistical methods: The geometric approach.
            Springer Science & Business Media, 2012.
    """
    alpha = 1 - conf
    df = result.model.data.frame
    groupby = df.groupby(factor)[value]
    cols = ['count', 'mean']
    groups = groupby.agg(cols)
    groups.columns = cols
    cis = _groups_ci(result, value, factor, conf)
    groups.loc[:, ['lower', 'upper']] = cis
    groups = groups.reindex(df.loc[:, factor].unique())

    if formatted:
        format_map = {
            'count': '{:.0f}',
            'mean': '{:,.4g}',
            'lower': '{:.4g}',
            'upper': '{:.4g}',
        }
        return groups.style.format(format_map)
    else:
        return groups

def interactions(df: pd.DataFrame, *, value: str, between: list[str], formatted=True):
    """
    Interaction table.

    Parameters
    ----------
    df : pandas.DataFrame
        Dataframe of measurements and categories.
    value : str
        Name of the column of measurements.
    between : list[str]
        List of columns to group over before averaging.
    formatted : bool, optional
        When True, returns a formatted DataFrame. Set this to False to access
        the calculated values.

    Returns
    -------
    pandas.DataFrame or pandas.io.formats.style.Styler
        Average values per pair of columns in between. A DataFrame when `formatted`
        is True, a Styler otherwise.
    """
    result = df.groupby([*between])[value].mean().unstack()

    if formatted:
        return result.style.format(precision=4)
    else:
        return result

################################################################################
## Residual analysis
################################################################################

def adequacy(result, formatted=True):
    """
    Table of statistics from a fitted OLS regression (see Returns).

    Parameters
    ----------
    result : statsmodels.regression.linear_model.RegressionResults
        The result of calling `fit` on a statsmodels linear regression model.
    formatted : bool, optional
        When True, returns a formatted DataFrame. Set this to False to access
        the calculated values.

    Returns
    -------
    pandas.DataFrame or pandas.io.formats.style.Styler
        A DataFrame when `formatted` is True, a Styler otherwise. A list of
        statistics from the regression with columns:

        S
            square-root of the mean-squared error
        R-sq
            coefficient of determination
        R-sq (adj)
            coefficient of determination, adjusted for degrees of freedom in model
        F
            F-statistic for the whole model
        PR(>F)
            p-value of the F-statistic
        AIC
            Akaike information criterion
        BIC
            Bayesian information criterion
        N
            number of observations
    """
    data = {
        'S': np.sqrt(result.mse_resid),
        'R-sq': result.rsquared,
        'R-sq (adj)': result.rsquared_adj,
        'F': result.fvalue,
        'PR(>F)': result.f_pvalue,
        'AIC': result.aic,
        'BIC': result.bic,
        'N': int(result.nobs),
    }
    df = pd.DataFrame(
        data,
        index=[''])

    if formatted:
        format_map = {
            'S': '{:.4g}',
            'R-sq': '{:.4g}',
            'R-sq (adj)': '{:.4g}',
            'F': '{:.2f}',
            'PR(>F)': '{:.3f}',
            'AIC': '{:.4g}',
            'BIC': '{:.4g}',
            'N': '{:.0f}',
        }
        return df.style.format(format_map)
    else:
        return df

################################################################################
## Confidence intervals
################################################################################

def _groups_ci(result, value, factor, conf):
    alpha = 1 - conf
    df = result.model.data.frame

    nobs = len(df)
    se_dist = st.t(result.df_resid)
    level_names = df.loc[:, factor].unique()
    n_levels = len(level_names)
    se = np.sqrt(result.mse_resid)
    groups = df.groupby(factor)[value]

    ci = np.zeros([n_levels, 2])
    for i, level in enumerate(level_names):
        nobs_level = groups.count().loc[level]
        half_width = se_dist.ppf(1 - alpha / 2) * se / np.sqrt(nobs_level)
        ci[i, :] = groups.mean().loc[level] + np.array([-half_width, half_width])
    return ci
